"""
ComfyUI Manager Server

Main server implementation providing REST API endpoints for ComfyUI Manager functionality.
Handles task queue management, custom node operations, model installation, and system configuration.
"""

import asyncio
import copy
import heapq
import json
import logging
import os
import platform
import re
import shutil
import subprocess  # don't remove this
import sys
import threading
import traceback
import urllib.request
import uuid
import zipfile
from datetime import datetime, timedelta
from typing import Any, Optional

import folder_paths
import latent_preview
import nodes
from aiohttp import web
from comfy.cli_args import args
from pydantic import ValidationError

from comfyui_manager.glob.utils import (
    formatting_utils,
    model_utils,
    security_utils,
    node_pack_utils,
    environment_utils,
)


from server import PromptServer

from . import manager_core as core
from ..common import manager_util
from ..common import cm_global
from ..common import manager_downloader
from ..common import context



from ..data_models import (
    QueueTaskItem,
    TaskHistoryItem,
    TaskStateMessage,
    TaskExecutionStatus,
    MessageTaskDone,
    MessageTaskStarted,
    MessageUpdate,
    ManagerMessageName,
    BatchExecutionRecord,
    ComfyUISystemState,
    ImportFailInfoBulkRequest,
    ImportFailInfoBulkResponse,
    BatchOperation,
    InstalledNodeInfo,
    ComfyUIVersionInfo,
    InstallPackParams,
    UpdatePackParams,
    UpdateComfyUIParams,
    FixPackParams,
    UninstallPackParams,
    DisablePackParams,
    EnablePackParams,
    ModelMetadata,
    OperationType,
    OperationResult,
    ManagerDatabaseSource,
    SecurityLevel,
    UpdateAllQueryParams,
    UpdateComfyUIQueryParams,
    ComfyUISwitchVersionQueryParams,
)

from .constants import (
    model_dir_name_map,
    SECURITY_MESSAGE_MIDDLE,
    SECURITY_MESSAGE_MIDDLE_P,
)

if not manager_util.is_manager_pip_package():
    network_mode_description = "offline"
else:
    network_mode_description = core.get_config()["network_mode"]
logging.info("[ComfyUI-Manager] network_mode: " + network_mode_description)


MAXIMUM_HISTORY_SIZE = 10000
routes = PromptServer.instance.routes


def is_loopback(address):
    import ipaddress

    try:
        return ipaddress.ip_address(address).is_loopback
    except ValueError:
        return False


def error_response(
    status: int, message: str, error_type: Optional[str] = None
) -> web.Response:
    """Create a standardized error response.

    Args:
        status: HTTP status code
        message: Error message
        error_type: Optional error type/category

    Returns:
        web.Response with JSON error body
    """
    error_data = {"error": message}
    if error_type:
        error_data["error_type"] = error_type

    return web.json_response(error_data, status=status)


class ManagerFuncsInComfyUI(core.ManagerFuncs):
    def get_current_preview_method(self):
        if args.preview_method == latent_preview.LatentPreviewMethod.Auto:
            return "auto"
        elif args.preview_method == latent_preview.LatentPreviewMethod.Latent2RGB:
            return "latent2rgb"
        elif args.preview_method == latent_preview.LatentPreviewMethod.TAESD:
            return "taesd"
        else:
            return "none"

    def run_script(self, cmd, cwd="."):
        if len(cmd) > 0 and cmd[0].startswith("#"):
            logging.error(f"[ComfyUI-Manager] Unexpected behavior: `{cmd}`")
            return 0

        process = subprocess.Popen(
            cmd,
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            bufsize=1,
            env=core.get_script_env(),
        )

        stdout_thread = threading.Thread(
            target=formatting_utils.handle_stream, args=(process.stdout, "")
        )
        stderr_thread = threading.Thread(
            target=formatting_utils.handle_stream, args=(process.stderr, "[!]")
        )

        stdout_thread.start()
        stderr_thread.start()

        stdout_thread.join()
        stderr_thread.join()

        return process.wait()


core.manager_funcs = ManagerFuncsInComfyUI()

from comfyui_manager.common.manager_downloader import (
    download_url,
    download_url_with_agent,
)


class TaskQueue:
    instance = None

    def __init__(self):
        TaskQueue.instance = self
        self.mutex = threading.RLock()
        self.not_empty = threading.Condition(self.mutex)
        self.current_index = 0
        self.pending_tasks = []
        self.running_tasks = {}
        self.history_tasks = {}
        self.task_counter = 0
        self.batch_id = None
        self.batch_start_time = None
        self.batch_state_before = None
        self._worker_task = None
        self._cleanup_performed = False

    def is_processing(self) -> bool:
        """Check if the queue is currently processing tasks"""
        return self._worker_task is not None and self._worker_task.is_alive()

    def start_worker(self) -> bool:
        """Start the task worker if not already running. Returns True if started, False if already running."""
        if self._worker_task is not None and self._worker_task.is_alive():
            logging.debug("[ComfyUI-Manager] Worker already running, skipping start")
            return False

        logging.debug("[ComfyUI-Manager] Starting task worker thread")
        self._worker_task = threading.Thread(target=lambda: asyncio.run(task_worker()))
        self._worker_task.start()
        return True

    def get_current_state(self) -> TaskStateMessage:
        return TaskStateMessage(
            history=self.get_history(),
            running_queue=self.get_current_queue()[0],
            pending_queue=self.get_current_queue()[1],
            installed_packs=core.get_installed_node_packs(),
        )

    @staticmethod
    def send_queue_state_update(
        msg: str, update: MessageUpdate, client_id: Optional[str] = None
    ) -> None:
        """Send queue state update to clients.

        Args:
            msg: Message type/event name
            update: Update data to send
            client_id: Optional client ID. If None, broadcasts to all clients.
                      If provided, sends only to that specific client.
        """
        PromptServer.instance.send_sync(msg, update.model_dump(mode="json"), client_id)

    def put(self, item) -> None:
        """Add a task to the queue. Item can be a dict or QueueTaskItem model."""
        with self.mutex:
            # Start a new batch if this is the first task after queue was empty
            if (
                self.batch_id is None
                and len(self.pending_tasks) == 0
                and len(self.running_tasks) == 0
            ):
                self._start_new_batch()

            # Convert to Pydantic model if it's a dict
            if isinstance(item, dict):
                item = QueueTaskItem(**item)

            # Use current_index as priority (earlier tasks have lower numbers)
            priority = self.current_index
            self.current_index += 1

            # Push tuple: (priority, task_counter, item)
            # task_counter ensures stable sort for items with same priority
            heapq.heappush(self.pending_tasks, (priority, self.task_counter, item))
            logging.debug(
                "[ComfyUI-Manager] Task added to queue: kind=%s, ui_id=%s, client_id=%s, pending_count=%d",
                item.kind,
                item.ui_id,
                item.client_id,
                len(self.pending_tasks),
            )
            self.not_empty.notify()

    def _start_new_batch(self) -> None:
        """Start a new batch session for tracking operations."""
        self.batch_id = (
            f"batch_{datetime.now().strftime('%Y%m%d_%H%M%S')}_{uuid.uuid4().hex[:8]}"
        )
        self.batch_start_time = datetime.now().isoformat()
        self.batch_state_before = self._capture_system_state()
        logging.debug("[ComfyUI-Manager] Started new batch: %s", self.batch_id)

    def get(
        self, timeout: Optional[float] = None
    ) -> tuple[Optional[QueueTaskItem], int]:
        with self.not_empty:
            while len(self.pending_tasks) == 0:
                self.not_empty.wait(timeout=timeout)
                if timeout is not None and len(self.pending_tasks) == 0:
                    logging.debug("[ComfyUI-Manager] Task queue get timed out")
                    return None
            # Pop tuple and extract the item
            priority, counter, item = heapq.heappop(self.pending_tasks)
            task_index = self.task_counter
            self.running_tasks[task_index] = copy.deepcopy(item)
            self.task_counter += 1
            logging.debug(
                "[ComfyUI-Manager] Task retrieved from queue: kind=%s, ui_id=%s, task_index=%d, running_count=%d, pending_count=%d",
                item.kind,
                item.ui_id,
                task_index,
                len(self.running_tasks),
                len(self.pending_tasks),
            )
            TaskQueue.send_queue_state_update(
                ManagerMessageName.cm_task_started.value,
                MessageTaskStarted(
                    ui_id=item.ui_id,
                    kind=item.kind,
                    timestamp=datetime.now(),
                    state=self.get_current_state(),
                ),
                client_id=item.client_id,  # Send task started only to the client that requested it
            )
            return item, task_index

    async def task_done(
        self,
        item: QueueTaskItem,
        task_index: int,
        result_msg: str,
        status: Optional[TaskExecutionStatus] = None,
    ) -> None:
        """Mark task as completed and add to history"""

        with self.mutex:
            now = datetime.now()
            timestamp = now.isoformat()

            # Remove task from running_tasks using the task_index
            self.running_tasks.pop(task_index, None)
            logging.debug(
                "[ComfyUI-Manager] Task completed: kind=%s, ui_id=%s, task_index=%d, status=%s, running_count=%d",
                item.kind,
                item.ui_id,
                task_index,
                status.status_str if status else "unknown",
                len(self.running_tasks),
            )

            # Manage history size
            if len(self.history_tasks) > MAXIMUM_HISTORY_SIZE:
                self.history_tasks.pop(next(iter(self.history_tasks)))

            # Update history
            self.history_tasks[item.ui_id] = TaskHistoryItem(
                ui_id=item.ui_id,
                client_id=item.client_id,
                timestamp=now,
                result=result_msg,
                kind=item.kind,
                status=status,
                batch_id=self.batch_id,
                end_time=now,
            )

        # Force cache refresh for successful pack-modifying operations
        pack_modifying_tasks = {
            OperationType.install.value,
            OperationType.uninstall.value,
            OperationType.enable.value,
            OperationType.disable.value,
        }
        if (
            item.kind in pack_modifying_tasks
            and status
            and status.status_str == OperationResult.success.value
        ):
            try:
                logging.debug(
                    "[ComfyUI-Manager] Refreshing cache after successful %s operation",
                    item.kind,
                )
                # Force unified_manager to refresh its installed packages cache
                await core.unified_manager.reload(
                    ManagerDatabaseSource.cache.value,
                    dont_wait=True,
                    update_cnr_map=False,
                )
            except Exception as e:
                logging.warning(
                    f"[ComfyUI-Manager] Failed to refresh cache after {item.kind}: {e}"
                )

        # Send WebSocket message indicating task is complete
        TaskQueue.send_queue_state_update(
            ManagerMessageName.cm_task_completed.value,
            MessageTaskDone(
                ui_id=item.ui_id,
                result=result_msg,
                kind=item.kind,
                status=status,
                timestamp=datetime.fromisoformat(timestamp),
                state=self.get_current_state(),
            ),
            client_id=item.client_id,  # Send completion only to the client that requested it
        )

    def get_current_queue(self) -> tuple[list[QueueTaskItem], list[QueueTaskItem]]:
        """Get current running and remaining tasks"""
        with self.mutex:
            running = list(self.running_tasks.values())
            # Extract items from tuples, maintaining heap order
            remaining = [item for priority, counter, item in sorted(self.pending_tasks)]
            return running, remaining

    def get_tasks_remaining(self) -> int:
        """Get number of tasks remaining"""
        with self.mutex:
            return len(self.pending_tasks) + len(self.running_tasks)

    def wipe_queue(self) -> None:
        """Clear all task queue"""
        with self.mutex:
            pending_count = len(self.pending_tasks)
            self.pending_tasks = []
            logging.debug(
                "[ComfyUI-Manager] Queue wiped: cleared %d pending tasks", pending_count
            )

    def abort(self) -> None:
        """Abort current operations"""
        with self.mutex:
            pending_count = len(self.pending_tasks)
            running_count = len(self.running_tasks)
            self.pending_tasks = []
            self.running_tasks = {}
            logging.debug(
                "[ComfyUI-Manager] Queue aborted: cleared %d pending and %d running tasks",
                pending_count,
                running_count,
            )

    def delete_history_item(self, ui_id: str) -> None:
        """Delete specific task from history"""
        with self.mutex:
            self.history_tasks.pop(ui_id, None)

    def get_history(
        self,
        ui_id: Optional[str] = None,
        max_items: Optional[int] = None,
        offset: int = -1,
    ) -> dict[str, TaskHistoryItem]:
        """Get task history. If ui_id (task id) is passsed, only return that task's history item entry."""
        with self.mutex:
            if ui_id is None:
                out = {}
                i = 0
                if offset < 0 and max_items is not None:
                    offset = len(self.history_tasks) - max_items
                for k in self.history_tasks:
                    if i >= offset:
                        out[k] = self.history_tasks[k]
                        if max_items is not None and len(out) >= max_items:
                            break
                    i += 1
                return out
            elif ui_id in self.history_tasks:
                return self.history_tasks[ui_id]
            else:
                return {}

    def done_count(self) -> int:
        """Get the number of completed tasks in history.

        Returns:
            int: Number of tasks that have been completed and are stored in history.
                 Returns 0 if history_tasks is None (defensive programming).
        """
        return len(self.history_tasks) if self.history_tasks is not None else 0

    def total_count(self) -> int:
        """Get the total number of tasks currently in the system (pending + running).

        Returns:
            int: Combined count of pending and running tasks.
                 Returns 0 if either collection is None (defensive programming).
        """
        return (
            len(self.pending_tasks) + len(self.running_tasks)
            if self.pending_tasks is not None and self.running_tasks is not None
            else 0
        )

    def finalize(self) -> None:
        """Finalize a completed task batch by saving execution history to disk.

        This method is intended to be called when the queue transitions from having
        tasks to being completely empty (no pending or running tasks). It will create
        a comprehensive snapshot of the ComfyUI state and all operations performed.
        """
        if self.batch_id is not None:
            batch_path = os.path.join(
                context.manager_batch_history_path, self.batch_id + ".json"
            )
            logging.debug(
                "[ComfyUI-Manager] Finalizing batch: batch_id=%s, history_count=%d",
                self.batch_id,
                len(self.history_tasks),
            )

            try:
                end_time = datetime.now().isoformat()
                state_after = self._capture_system_state()
                operations = self._extract_batch_operations()

                batch_record = BatchExecutionRecord(
                    batch_id=self.batch_id,
                    start_time=self.batch_start_time,
                    end_time=end_time,
                    state_before=self.batch_state_before,
                    state_after=state_after,
                    operations=operations,
                    total_operations=len(operations),
                    successful_operations=len(
                        [
                            op
                            for op in operations
                            if op.result == OperationResult.success.value
                        ]
                    ),
                    failed_operations=len(
                        [
                            op
                            for op in operations
                            if op.result == OperationResult.failed.value
                        ]
                    ),
                    skipped_operations=len(
                        [
                            op
                            for op in operations
                            if op.result == OperationResult.skipped.value
                        ]
                    ),
                )

                # Save to disk
                with open(batch_path, "w", encoding="utf-8") as json_file:
                    json.dump(
                        batch_record.model_dump(), json_file, indent=4, default=str
                    )

                logging.debug(
                    "[ComfyUI-Manager] Batch history saved: batch_id=%s, path=%s, total_ops=%d, successful=%d, failed=%d, skipped=%d",
                    self.batch_id,
                    batch_path,
                    batch_record.total_operations,
                    batch_record.successful_operations,
                    batch_record.failed_operations,
                    batch_record.skipped_operations,
                )

                # Reset batch tracking
                self.batch_id = None
                self.batch_start_time = None
                self.batch_state_before = None

                # Cleanup old batch files once per session
                if not self._cleanup_performed:
                    self._cleanup_old_batches()
                    self._cleanup_performed = True

            except Exception as e:
                logging.error(f"[ComfyUI-Manager] Failed to save batch history: {e}")

    def _capture_system_state(self) -> ComfyUISystemState:
        """Capture current ComfyUI system state for batch record."""
        logging.debug("[ComfyUI-Manager] Capturing system state for batch record")
        return ComfyUISystemState(
            snapshot_time=datetime.now().isoformat(),
            comfyui_version=self._get_comfyui_version_info(),
            frontend_version=self._get_frontend_version(),
            python_version=platform.python_version(),
            platform_info=f"{platform.system()} {platform.release()} ({platform.machine()})",
            installed_nodes=self._get_installed_nodes(),
            comfyui_root_path=self._get_comfyui_root_path(),
            model_paths=self._get_model_paths(),
            manager_version=self._get_manager_version(),
            security_level=self._get_security_level(),
            network_mode=self._get_network_mode(),
            cli_args=self._get_cli_args(),
            custom_nodes_count=self._get_custom_nodes_count(),
            failed_imports=self._get_failed_imports(),
            pip_packages=self._get_pip_packages(),
            manager_config=core.get_config(),
            embedded_python=os.path.split(os.path.split(sys.executable)[0])[1] == "python_embeded",
        )

    def _get_comfyui_version_info(self) -> ComfyUIVersionInfo:
        """Get ComfyUI version information."""
        try:
            version_info = core.get_comfyui_versions()
            current_version = version_info[1] if len(version_info) > 1 else "unknown"
            return ComfyUIVersionInfo(version=current_version)
        except Exception:
            return ComfyUIVersionInfo(version="unknown")

    def _get_frontend_version(self) -> Optional[str]:
        """Get ComfyUI frontend version."""
        try:
            # Check if front-end-root is specified (overrides version)
            if hasattr(args, "front_end_root") and args.front_end_root:
                return f"custom-root: {args.front_end_root}"

            # Check if front-end-version is specified
            if hasattr(args, "front_end_version") and args.front_end_version:
                if "@" in args.front_end_version:
                    return args.front_end_version.split("@")[1]
                else:
                    return args.front_end_version

            # Otherwise, check for installed package
            pip_packages = self._get_pip_packages()
            for package_name in ["comfyui-frontend", "comfyui_frontend"]:
                if package_name in pip_packages:
                    return pip_packages[package_name]

            return None
        except Exception:
            return None

    def _get_installed_nodes(self) -> dict[str, InstalledNodeInfo]:
        """Get information about installed node packages."""
        installed_nodes = {}

        try:
            node_packs = core.get_installed_node_packs()
            for pack_name, pack_info in node_packs.items():
                # Determine install method and repository URL
                install_method = "git" if pack_info.get("aux_id") else "cnr"
                repository_url = None

                if pack_info.get("aux_id"):
                    # It's a git-based node, construct GitHub URL
                    repository_url = f"https://github.com/{pack_info['aux_id']}"

                installed_nodes[pack_name] = InstalledNodeInfo(
                    name=pack_name,
                    version=pack_info.get("ver", "unknown"),
                    install_method=install_method,
                    repository_url=repository_url,
                    enabled=pack_info.get("enabled", True),
                )
        except Exception as e:
            logging.warning(f"[ComfyUI-Manager] Failed to get installed nodes: {e}")

        return installed_nodes

    def _get_comfyui_root_path(self) -> str:
        """Get ComfyUI root installation directory."""
        try:
            return os.path.dirname(folder_paths.__file__)
        except Exception:
            return None

    def _get_model_paths(self) -> dict[str, list[str]]:
        """Get model paths for different model types."""
        try:
            model_paths = {}
            for model_type in model_dir_name_map.keys():
                try:
                    paths = folder_paths.get_folder_paths(model_type)
                    if paths:
                        model_paths[model_type] = paths
                except Exception:
                    continue
            return model_paths
        except Exception:
            return {}

    def _get_manager_version(self) -> str:
        """Get ComfyUI Manager version."""
        try:
            version_code = getattr(core, "version_code", [4, 0])
            return f"V{version_code[0]}.{version_code[1]}"
        except Exception:
            return None

    def _get_security_level(self) -> SecurityLevel:
        """Get current security level."""
        try:
            config = core.get_config()
            level_str = config.get("security_level", "normal")
            # Map the string to SecurityLevel enum
            level_mapping = {
                "strong": SecurityLevel.strong,
                "normal": SecurityLevel.normal,
                "normal-": SecurityLevel.normal_,
                "weak": SecurityLevel.weak,
            }
            return level_mapping.get(level_str, SecurityLevel.normal)
        except Exception:
            return None

    def _get_network_mode(self) -> str:
        """Get current network mode."""
        try:
            config = core.get_config()
            return config.get("network_mode", "online")
        except Exception:
            return None

    def _get_cli_args(self) -> dict[str, Any]:
        """Get selected CLI arguments."""
        try:
            cli_args = {}
            if hasattr(args, "listen"):
                cli_args["listen"] = args.listen
            if hasattr(args, "port"):
                cli_args["port"] = args.port
            if hasattr(args, "preview_method"):
                cli_args["preview_method"] = str(args.preview_method)
            if hasattr(args, "enable_manager_legacy_ui"):
                cli_args["enable_manager_legacy_ui"] = args.enable_manager_legacy_ui
            if hasattr(args, "front_end_version"):
                cli_args["front_end_version"] = args.front_end_version
            if hasattr(args, "front_end_root"):
                cli_args["front_end_root"] = args.front_end_root
            return cli_args
        except Exception:
            return {}

    def _get_custom_nodes_count(self) -> int:
        """Get total number of custom node packages."""
        try:
            node_packs = core.get_installed_node_packs()
            return len(node_packs)
        except Exception:
            return 0

    def _get_failed_imports(self) -> list[str]:
        """Get list of custom nodes that failed to import."""
        try:
            # Check if the import_failed_extensions set is available
            if hasattr(sys, "__comfyui_manager_import_failed_extensions"):
                failed_set = getattr(sys, "__comfyui_manager_import_failed_extensions")
                return list(failed_set) if failed_set else []
            return []
        except Exception:
            return []

    def _get_pip_packages(self) -> dict[str, str]:
        """Get installed pip packages."""
        try:
            return core.get_installed_pip_packages()
        except Exception:
            return {}

    def _extract_batch_operations(self) -> list[BatchOperation]:
        """Extract operations from completed task history for this batch."""
        operations = []

        try:
            for ui_id, task in self.history_tasks.items():
                # Only include operations from the current batch
                if task.batch_id != self.batch_id:
                    continue

                result_status = OperationResult.success
                if task.status:
                    status_str = (
                        task.status.status_str
                        if hasattr(task.status, "status_str")
                        else task.status.get(
                            "status_str", OperationResult.success.value
                        )
                    )
                    if status_str == OperationResult.error.value:
                        result_status = OperationResult.failed
                    elif status_str == OperationResult.skip.value:
                        result_status = OperationResult.skipped

                operation = BatchOperation(
                    operation_id=ui_id,
                    operation_type=task.kind,
                    target=f"task_{ui_id}",
                    result=result_status.value,
                    start_time=task.timestamp,
                    end_time=task.end_time,
                    client_id=task.client_id,
                )
                operations.append(operation)
        except Exception as e:
            logging.warning(
                f"[ComfyUI-Manager] Failed to extract batch operations: {e}"
            )

        return operations

    def _cleanup_old_batches(self) -> None:
        """Clean up batch history files older than 90 days.

        This is a best-effort cleanup that silently ignores any errors
        to avoid disrupting normal operations.
        """
        try:
            cutoff = datetime.now() - timedelta(days=16)
            cutoff_timestamp = cutoff.timestamp()

            pattern = os.path.join(context.manager_batch_history_path, "batch_*.json")
            removed_count = 0

            import glob

            for file_path in glob.glob(pattern):
                try:
                    if os.path.getmtime(file_path) < cutoff_timestamp:
                        os.remove(file_path)
                        removed_count += 1
                except Exception:
                    pass

            if removed_count > 0:
                logging.debug(
                    "[ComfyUI-Manager] Cleaned up %d old batch history files",
                    removed_count,
                )

        except Exception:
            # Silently ignore all errors - this is non-critical functionality
            pass


task_queue = TaskQueue()

# Preview method initialization
if args.preview_method == latent_preview.LatentPreviewMethod.NoPreviews:
    environment_utils.set_preview_method(core.get_config()["preview_method"])
else:
    logging.warning(
        "[ComfyUI-Manager] Since --preview-method is set, ComfyUI-Manager's preview method feature will be ignored."
    )


async def task_worker():
    logging.debug("[ComfyUI-Manager] Task worker started")
    await core.unified_manager.reload(ManagerDatabaseSource.cache.value)

    async def do_install(params: InstallPackParams) -> str:
        if not security_utils.is_allowed_security_level('middle+'):
            logging.error(SECURITY_MESSAGE_MIDDLE_P)
            return OperationResult.failed.value

        node_id = params.id
        node_version = params.selected_version
        channel = params.channel
        mode = params.mode
        skip_post_install = params.skip_post_install

        logging.debug(
            "[ComfyUI-Manager] Installing node: id=%s, version=%s, channel=%s, mode=%s",
            node_id,
            node_version,
            channel,
            mode,
        )

        try:
            node_spec = core.unified_manager.resolve_node_spec(
                f"{node_id}@{node_version}"
            )
            if node_spec is None:
                logging.error(
                    f"Cannot resolve install target: '{node_id}@{node_version}'"
                )
                return f"Cannot resolve install target: '{node_id}@{node_version}'"

            node_name, version_spec, is_specified = node_spec
            res = await core.unified_manager.install_by_id(
                node_name,
                version_spec,
                channel,
                mode,
                return_postinstall=skip_post_install,
            )  # discard post install if skip_post_install mode

            if res.action not in [
                "skip",
                "enable",
                "install-git",
                "install-cnr",
                "switch-cnr",
            ]:
                logging.error(f"[ComfyUI-Manager] Installation failed:\n{res.msg}")
                return res.msg

            elif not res.result:
                logging.error(f"[ComfyUI-Manager] Installation failed:\n{res.msg}")
                return res.msg

            return OperationResult.success.value
        except Exception:
            traceback.print_exc()
            return "Installation failed"

    async def do_enable(params: EnablePackParams) -> str:
        cnr_id = params.cnr_id
        logging.debug("[ComfyUI-Manager] Enabling node: cnr_id=%s", cnr_id)
        core.unified_manager.unified_enable(cnr_id)
        return OperationResult.success.value

    async def do_update(params: UpdatePackParams) -> dict[str, str]:
        node_name = params.node_name
        node_ver = params.node_ver

        logging.debug(
            "[ComfyUI-Manager] Updating node: name=%s, version=%s", node_name, node_ver
        )

        try:
            res = core.unified_manager.unified_update(node_name, node_ver)

            if res.ver == "unknown":
                url = core.unified_manager.unknown_active_nodes[node_name][0]
                try:
                    title = os.path.basename(url)
                except Exception:
                    title = node_name
            else:
                url = core.unified_manager.cnr_map[node_name].get("repository")
                title = core.unified_manager.cnr_map[node_name]["name"]

            manager_util.clear_pip_cache()

            if url is not None:
                base_res = {"url": url, "title": title}
            else:
                base_res = {"title": title}

            if res.result:
                if res.action == "skip":
                    base_res["msg"] = OperationResult.skip.value
                    return base_res
                else:
                    base_res["msg"] = OperationResult.success.value
                    return base_res

            base_res["msg"] = f"An error occurred while updating '{node_name}'."
            logging.error(
                f"\nERROR: An error occurred while updating '{node_name}'. (res.result={res.result}, res.action={res.action})"
            )
            return base_res
        except Exception:
            traceback.print_exc()

        return {"msg": f"An error occurred while updating '{node_name}'."}

    async def do_update_comfyui(params: UpdateComfyUIParams) -> str:
        try:
            repo_path = os.path.dirname(folder_paths.__file__)

            # Check if this is a version switch operation
            if params.target_version:
                # Switch to specific version
                logging.info(f"Switching ComfyUI to version: {params.target_version}")
                core.switch_comfyui(params.target_version)
                return f"success-switched-{params.target_version}"
            else:
                # Regular update operation
                is_stable = params.is_stable if params.is_stable is not None else True
                logging.debug(
                    "[ComfyUI-Manager] Updating ComfyUI: is_stable=%s, repo_path=%s",
                    is_stable,
                    repo_path,
                )
                latest_tag = None
                if is_stable:
                    res, latest_tag = core.update_to_stable_comfyui(repo_path)
                else:
                    res = core.update_path(repo_path)

                if res == "fail":
                    logging.error("ComfyUI update failed")
                    return "fail"
                elif res == "updated":
                    if is_stable:
                        logging.info("ComfyUI is updated to latest stable version.")
                        return "success-stable-" + latest_tag
                    else:
                        logging.info("ComfyUI is updated to latest nightly version.")
                        return "success-nightly"
                else:  # skipped
                    logging.info("ComfyUI is up-to-date.")
                    return OperationResult.skip.value

        except Exception:
            traceback.print_exc()

        return "An error occurred while updating 'comfyui'."

    async def do_fix(params: FixPackParams) -> str:
        if not security_utils.is_allowed_security_level('middle'):
            logging.error(SECURITY_MESSAGE_MIDDLE)
            return OperationResult.failed.value

        node_name = params.node_name
        node_ver = params.node_ver

        try:
            res = core.unified_manager.unified_fix(node_name, node_ver)

            if res.result:
                return OperationResult.success.value
            else:
                logging.error(res.msg)

            logging.error(
                f"\nERROR: An error occurred while fixing '{node_name}@{node_ver}'."
            )
        except Exception:
            traceback.print_exc()

        return f"An error occurred while fixing '{node_name}@{node_ver}'."

    async def do_uninstall(params: UninstallPackParams) -> str:
        if not security_utils.is_allowed_security_level('middle'):
            logging.error(SECURITY_MESSAGE_MIDDLE)
            return OperationResult.failed.value

        node_name = params.node_name
        is_unknown = params.is_unknown

        logging.debug(
            "[ComfyUI-Manager] Uninstalling node: name=%s, is_unknown=%s",
            node_name,
            is_unknown,
        )

        try:
            res = core.unified_manager.unified_uninstall(node_name, is_unknown)

            if res.result:
                return OperationResult.success.value

            logging.error(
                f"\nERROR: An error occurred while uninstalling '{node_name}'."
            )
        except Exception:
            traceback.print_exc()

        return f"An error occurred while uninstalling '{node_name}'."

    async def do_disable(params: DisablePackParams) -> str:
        node_name = params.node_name

        logging.debug(
            "[ComfyUI-Manager] Disabling node: name=%s, is_unknown=%s",
            node_name,
            params.is_unknown,
        )

        try:
            res = core.unified_manager.unified_disable(node_name, params.is_unknown)

            if res:
                return OperationResult.success.value

        except Exception:
            traceback.print_exc()

        return f"Failed to disable: '{node_name}'"

    async def do_install_model(params: ModelMetadata) -> str:
        if not security_utils.is_allowed_security_level('middle+'):
            logging.error(SECURITY_MESSAGE_MIDDLE_P)
            return OperationResult.failed.value

        json_data = params.model_dump()

        model_path = model_utils.get_model_path(json_data)
        model_url = json_data.get("url")

        res = False

        try:
            if model_path is not None:
                logging.info(
                    f"Install model '{json_data['name']}' from '{model_url}' into '{model_path}'"
                )

                if json_data["filename"] == "<huggingface>":
                    if os.path.exists(
                        os.path.join(model_path, os.path.dirname(json_data["url"]))
                    ):
                        logging.error(
                            f"[ComfyUI-Manager] the model path already exists: {model_path}"
                        )
                        return f"The model path already exists: {model_path}"

                    logging.info(
                        f"[ComfyUI-Manager] Downloading '{model_url}' into '{model_path}'"
                    )
                    manager_downloader.download_repo_in_bytes(
                        repo_id=model_url, local_dir=model_path
                    )

                    return OperationResult.success.value

                elif not core.get_config()["model_download_by_agent"] and (
                    model_url.startswith("https://github.com")
                    or model_url.startswith("https://huggingface.co")
                    or model_url.startswith("https://heibox.uni-heidelberg.de")
                ):
                    model_dir = model_utils.get_model_dir(json_data, True)
                    download_url(model_url, model_dir, filename=json_data["filename"])
                    if model_path.endswith(".zip"):
                        res = core.unzip(model_path)
                    else:
                        res = True

                    if res:
                        return OperationResult.success.value
                else:
                    res = download_url_with_agent(model_url, model_path)
                    if res and model_path.endswith(".zip"):
                        res = core.unzip(model_path)
            else:
                logging.error(
                    f"[ComfyUI-Manager] Model installation error: invalid model type - {json_data['type']}"
                )

            if res:
                return OperationResult.success.value

        except Exception as e:
            logging.error(f"[ComfyUI-Manager] ERROR: {e}")

        return f"Model installation error: {model_url}"

    while True:
        timeout = 4.0
        task = task_queue.get(timeout)
        if task is None:
            is_empty_queue = (
                task_queue.total_count() == 0 and len(task_queue.running_tasks) == 0
            )
            if is_empty_queue:
                logging.debug("[ComfyUI-Manager] Queue empty - all tasks completed")

                did_complete_tasks = task_queue.done_count() > 0
                if did_complete_tasks:
                    logging.debug(
                        "[ComfyUI-Manager] Finalizing batch history with %d completed tasks",
                        task_queue.done_count(),
                    )
                    task_queue.finalize()
                    logging.debug("[ComfyUI-Manager] Batch finalization complete")

                logging.info("\nAfter restarting ComfyUI, please refresh the browser.")

                res = {"status": "all-done"}

                # Broadcast general status updates to all clients
                logging.debug("[ComfyUI-Manager] Broadcasting queue all-done status")
                PromptServer.instance.send_sync("cm-queue-status", res)

                logging.debug("[ComfyUI-Manager] Task worker exiting")
                return

        item, task_index = task
        kind = item.kind

        logging.debug(
            "[ComfyUI-Manager] Processing task: kind=%s, ui_id=%s, client_id=%s, task_index=%d",
            kind,
            item.ui_id,
            item.client_id,
            task_index,
        )

        try:
            if kind == OperationType.install.value:
                msg = await do_install(item.params)
            elif kind == OperationType.enable.value:
                msg = await do_enable(item.params)
            elif kind == OperationType.install_model.value:
                msg = await do_install_model(item.params)
            elif kind == OperationType.update.value:
                msg = await do_update(item.params)
            elif kind == "update-main":
                msg = await do_update(item.params)
            elif kind == OperationType.update_comfyui.value:
                msg = await do_update_comfyui(item.params)
            elif kind == OperationType.fix.value:
                msg = await do_fix(item.params)
            elif kind == OperationType.uninstall.value:
                msg = await do_uninstall(item.params)
            elif kind == OperationType.disable.value:
                msg = await do_disable(item.params)
            else:
                msg = "Unexpected kind: " + kind
        except Exception:
            msg = f"Exception: {(kind, item)}"
            logging.error(
                "[ComfyUI-Manager] Task execution exception: kind=%s, ui_id=%s, error=%s",
                kind,
                item.ui_id,
                traceback.format_exc(),
            )
            await task_queue.task_done(
                item,
                task_index,
                msg,
                TaskExecutionStatus(
                    status_str=OperationResult.error, completed=True, messages=[msg]
                ),
            )
            return

        # Determine status and message for task completion
        if isinstance(msg, dict) and "msg" in msg:
            result_msg = msg["msg"]
        else:
            result_msg = msg

        # Determine status
        if result_msg == OperationResult.success.value:
            status = TaskExecutionStatus(
                status_str=OperationResult.success, completed=True, messages=[]
            )
        elif result_msg == OperationResult.skip.value:
            status = TaskExecutionStatus(
                status_str=OperationResult.skip, completed=True, messages=[]
            )
        else:
            status = TaskExecutionStatus(
                status_str=OperationResult.error, completed=True, messages=[result_msg]
            )

        logging.debug(
            "[ComfyUI-Manager] Task execution completed: kind=%s, ui_id=%s, status=%s, result=%s",
            kind,
            item.ui_id,
            status.status_str,
            result_msg,
        )
        await task_queue.task_done(item, task_index, result_msg, status)


@routes.post("/v2/manager/queue/task")
async def queue_task(request) -> web.Response:
    """Add a new task to the processing queue.

    Accepts task data via JSON POST and adds it to the TaskQueue for processing.
    The task worker will automatically pick up and process queued tasks.

    Args:
        request: aiohttp request containing JSON task data

    Returns:
        web.Response: HTTP 200 on successful queueing, HTTP 400 on validation error
    """
    try:
        json_data = await request.json()
        # Validate input using Pydantic model
        task_item = QueueTaskItem.model_validate(json_data)
        logging.debug(
            "[ComfyUI-Manager] Queueing task via API: kind=%s, ui_id=%s, client_id=%s",
            task_item.kind,
            task_item.ui_id,
            task_item.client_id,
        )
        TaskQueue.instance.put(task_item)
        # maybe start worker
        return web.Response(status=200)
    except ValidationError as e:
        logging.error(f"[ComfyUI-Manager] Invalid task data: {e}")
        return web.Response(status=400, text=f"Invalid task data: {e}")
    except Exception as e:
        logging.error(f"[ComfyUI-Manager] Error processing task: {e}")
        return web.Response(status=500, text="Internal server error")


@routes.get("/v2/manager/queue/history_list")
async def get_history_list(request) -> web.Response:
    """Get list of available batch history files.

    Returns a list of batch history IDs sorted by modification time (newest first).
    These IDs can be used with the history endpoint to retrieve detailed batch information.

    Returns:
        web.Response: JSON response with 'ids' array of history file IDs
    """
    history_path = context.manager_batch_history_path

    try:
        files = [
            os.path.join(history_path, f)
            for f in os.listdir(history_path)
            if os.path.isfile(os.path.join(history_path, f))
        ]
        files.sort(key=lambda x: os.path.getmtime(x), reverse=True)
        history_ids = [os.path.basename(f)[:-5] for f in files]

        return web.json_response(
            {"ids": list(history_ids)}, content_type="application/json"
        )
    except Exception as e:
        logging.error(f"[ComfyUI-Manager] /v2/manager/queue/history_list - {e}")
        return web.Response(status=400)


@routes.get("/v2/manager/queue/history")
async def get_history(request):
    """Get task history with optional client filtering.

    Query parameters:
        id: Batch history ID (for file-based history)
        client_id: Optional client ID to filter current session history
        ui_id: Optional specific task ID to get single task history
        max_items: Maximum number of items to return
        offset: Offset for pagination

    Returns:
        JSON with filtered history data
    """
    try:
        # Handle file-based batch history
        if "id" in request.rel_url.query:
            json_name = request.rel_url.query["id"] + ".json"
            batch_path = os.path.join(context.manager_batch_history_path, json_name)
            logging.debug(
                "[ComfyUI-Manager] Fetching batch history: id=%s",
                request.rel_url.query["id"],
            )

            with open(batch_path, "r", encoding="utf-8") as file:
                json_str = file.read()
                json_obj = json.loads(json_str)
                return web.json_response(json_obj, content_type="application/json")

        # Handle current session history with optional filtering
        client_id = request.rel_url.query.get("client_id")
        ui_id = request.rel_url.query.get("ui_id")
        max_items = request.rel_url.query.get("max_items")
        offset = request.rel_url.query.get("offset", -1)

        logging.debug(
            "[ComfyUI-Manager] Fetching history: client_id=%s, ui_id=%s, max_items=%s",
            client_id,
            ui_id,
            max_items,
        )

        if max_items:
            max_items = int(max_items)
        if offset:
            offset = int(offset)

        # Get history from TaskQueue
        if ui_id:
            history = task_queue.get_history(ui_id=ui_id)
        else:
            history = task_queue.get_history(max_items=max_items, offset=offset)

        # Filter by client_id if provided
        if client_id and isinstance(history, dict):
            filtered_history = {
                task_id: task_data
                for task_id, task_data in history.items()
                if hasattr(task_data, "client_id") and task_data.client_id == client_id
            }
            history = filtered_history

        return web.json_response({"history": history}, content_type="application/json")

    except Exception as e:
        logging.error(f"[ComfyUI-Manager] /v2/manager/queue/history - {e}")

    return web.Response(status=400)


@routes.get("/v2/customnode/getmappings")
async def fetch_customnode_mappings(request):
    """
    provide unified (node -> node pack) mapping list
    """
    mode = request.rel_url.query["mode"]

    nickname_mode = False
    if mode == "nickname":
        mode = "local"
        nickname_mode = True

    json_obj = await core.get_data_by_mode(mode, "extension-node-map.json")
    json_obj = core.map_to_unified_keys(json_obj)

    if nickname_mode:
        json_obj = node_pack_utils.nickname_filter(json_obj)

    all_nodes = set()
    patterns = []
    for k, x in json_obj.items():
        all_nodes.update(set(x[0]))

        if "nodename_pattern" in x[1]:
            patterns.append((x[1]["nodename_pattern"], x[0]))

    missing_nodes = set(nodes.NODE_CLASS_MAPPINGS.keys()) - all_nodes

    for x in missing_nodes:
        for pat, item in patterns:
            if re.match(pat, x):
                item.append(x)

    return web.json_response(json_obj, content_type="application/json")


@routes.get("/v2/customnode/fetch_updates")
async def fetch_updates(request):
    """
    DEPRECATED: This endpoint is no longer supported.

    Repository fetching has been removed from the API.
    Updates should be performed through the queue system using update operations.
    """
    return web.json_response(
        {
            "error": "This endpoint has been deprecated",
            "message": "Repository fetching is no longer supported. Please use the update operations through the queue system.",
            "deprecated": True,
        },
        status=410,  # 410 Gone
    )


@routes.get("/v2/manager/queue/update_all")
async def update_all(request: web.Request) -> web.Response:
    try:
        # Validate query parameters using Pydantic model
        query_params = UpdateAllQueryParams.model_validate(dict(request.rel_url.query))
        return await _update_all(query_params)
    except ValidationError as e:
        return web.json_response(
            {"error": "Validation error", "details": e.errors()}, status=400
        )


async def _update_all(params: UpdateAllQueryParams) -> web.Response:
    if not security_utils.is_allowed_security_level("middle+"):
        logging.error(SECURITY_MESSAGE_MIDDLE_P)
        return web.Response(status=403)

    # Extract client info from validated params
    base_ui_id = params.ui_id
    client_id = params.client_id
    mode = params.mode.value if params.mode else ManagerDatabaseSource.remote.value

    logging.debug(
        "[ComfyUI-Manager] Update all requested: client_id=%s, base_ui_id=%s, mode=%s",
        client_id,
        base_ui_id,
        mode,
    )

    if mode == ManagerDatabaseSource.local.value:
        channel = "local"
    else:
        channel = core.get_config()["channel_url"]

    await core.unified_manager.reload(mode)
    await core.unified_manager.get_custom_nodes(channel, mode)

    update_count = 0
    for k, v in core.unified_manager.active_nodes.items():
        if k == "comfyui-manager":
            # skip updating comfyui-manager if desktop version
            if os.environ.get("__COMFYUI_DESKTOP_VERSION__"):
                continue

        update_task = QueueTaskItem(
            kind=OperationType.update.value,
            ui_id=f"{base_ui_id}_{k}",  # Use client's base ui_id + node name
            client_id=client_id,
            params=UpdatePackParams(node_name=k, node_ver=v[0]),
        )
        task_queue.put(update_task)
        update_count += 1

    for k, v in core.unified_manager.unknown_active_nodes.items():
        if k == "comfyui-manager":
            # skip updating comfyui-manager if desktop version
            if os.environ.get("__COMFYUI_DESKTOP_VERSION__"):
                continue

        update_task = QueueTaskItem(
            kind=OperationType.update.value,
            ui_id=f"{base_ui_id}_{k}",  # Use client's base ui_id + node name
            client_id=client_id,
            params=UpdatePackParams(node_name=k, node_ver="unknown"),
        )
        task_queue.put(update_task)
        update_count += 1

    logging.debug(
        "[ComfyUI-Manager] Update all queued %d tasks for client_id=%s",
        update_count,
        client_id,
    )
    return web.Response(status=200)


@routes.get("/v2/manager/is_legacy_manager_ui")
async def is_legacy_manager_ui(request):
    return web.json_response(
        {"is_legacy_manager_ui": args.enable_manager_legacy_ui},
        content_type="application/json",
        status=200,
    )


# freeze imported version
startup_time_installed_node_packs = core.get_installed_node_packs()


@routes.get("/v2/customnode/installed")
async def installed_list(request):
    mode = request.query.get("mode", "default")

    if mode == "imported":
        res = startup_time_installed_node_packs
    else:
        res = core.get_installed_node_packs()

    return web.json_response(res, content_type="application/json")


@routes.get("/v2/snapshot/getlist")
async def get_snapshot_list(request):
    items = [
        f[:-5] for f in os.listdir(context.manager_snapshot_path) if f.endswith(".json")
    ]
    items.sort(reverse=True)
    return web.json_response({"items": items}, content_type="application/json")


@routes.get("/v2/snapshot/remove")
async def remove_snapshot(request):
    if not security_utils.is_allowed_security_level("middle"):
        logging.error(SECURITY_MESSAGE_MIDDLE)
        return web.Response(status=403)

    try:
        target = request.rel_url.query["target"]

        path = os.path.join(context.manager_snapshot_path, f"{target}.json")
        if os.path.exists(path):
            os.remove(path)

        return web.Response(status=200)
    except Exception:
        return web.Response(status=400)


@routes.get("/v2/snapshot/restore")
async def restore_snapshot(request):
    if not security_utils.is_allowed_security_level("middle+"):
        logging.error(SECURITY_MESSAGE_MIDDLE_P)
        return web.Response(status=403)

    try:
        target = request.rel_url.query["target"]

        path = os.path.join(context.manager_snapshot_path, f"{target}.json")
        if os.path.exists(path):
            if not os.path.exists(context.manager_startup_script_path):
                os.makedirs(context.manager_startup_script_path)

            target_path = os.path.join(
                context.manager_startup_script_path, "restore-snapshot.json"
            )
            shutil.copy(path, target_path)

            logging.info(f"Snapshot restore scheduled: `{target}`")
            return web.Response(status=200)

        logging.error(f"Snapshot file not found: `{path}`")
        return web.Response(status=400)
    except Exception:
        return web.Response(status=400)


@routes.get("/v2/snapshot/get_current")
async def get_current_snapshot_api(request):
    try:
        return web.json_response(
            await core.get_current_snapshot(), content_type="application/json"
        )
    except Exception:
        return web.Response(status=400)


@routes.get("/v2/snapshot/save")
async def save_snapshot(request):
    try:
        await core.save_snapshot_with_postfix("snapshot")
        return web.Response(status=200)
    except Exception:
        return web.Response(status=400)


def unzip_install(files):
    temp_filename = "manager-temp.zip"
    for url in files:
        if url.endswith("/"):
            url = url[:-1]
        try:
            headers = {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3"
            }

            req = urllib.request.Request(url, headers=headers)
            response = urllib.request.urlopen(req)
            data = response.read()

            with open(temp_filename, "wb") as f:
                f.write(data)

            with zipfile.ZipFile(temp_filename, "r") as zip_ref:
                zip_ref.extractall(core.get_default_custom_nodes_path())

            os.remove(temp_filename)
        except Exception as e:
            logging.error(f"Install(unzip) error: {url} / {e}")
            return False

    logging.info("Installation was successful.")
    return True


@routes.post("/v2/customnode/import_fail_info")
async def import_fail_info(request):
    try:
        json_data = await request.json()

        # Basic validation - ensure we have either cnr_id or url
        if not isinstance(json_data, dict):
            return web.Response(status=400, text="Request body must be a JSON object")

        if "cnr_id" not in json_data and "url" not in json_data:
            return web.Response(
                status=400, text="Either 'cnr_id' or 'url' field is required"
            )

        if "cnr_id" in json_data:
            if not isinstance(json_data["cnr_id"], str):
                return web.Response(status=400, text="'cnr_id' must be a string")
            module_name = core.unified_manager.get_module_name(json_data["cnr_id"])
        else:
            if not isinstance(json_data["url"], str):
                return web.Response(status=400, text="'url' must be a string")
            module_name = core.unified_manager.get_module_name(json_data["url"])

        if module_name is not None:
            info = cm_global.error_dict.get(module_name)
            if info is not None:
                return web.json_response(info)

        return web.Response(status=400)
    except Exception as e:
        logging.error(f"[ComfyUI-Manager] Error processing import fail info: {e}")
        return web.Response(status=500, text="Internal server error")


@routes.post("/v2/customnode/import_fail_info_bulk")
async def import_fail_info_bulk(request):
    try:
        json_data = await request.json()
        
        # Validate input using Pydantic model
        request_data = ImportFailInfoBulkRequest.model_validate(json_data)
        
        # Ensure we have either cnr_ids or urls
        if not request_data.cnr_ids and not request_data.urls:
            return web.Response(
                status=400, text="Either 'cnr_ids' or 'urls' field is required"
            )

        await core.unified_manager.reload('cache')
        await core.unified_manager.get_custom_nodes('default', 'cache')

        results = {}

        if request_data.cnr_ids:
            for cnr_id in request_data.cnr_ids:
                module_name = core.unified_manager.get_module_name(cnr_id)
                if module_name is not None:
                    info = cm_global.error_dict.get(module_name)
                    if info is not None:
                        results[cnr_id] = info
                    else:
                        results[cnr_id] = None
                else:
                    results[cnr_id] = None

        if request_data.urls:
            for url in request_data.urls:
                module_name = core.unified_manager.get_module_name(url)
                if module_name is not None:
                    info = cm_global.error_dict.get(module_name)
                    if info is not None:
                        results[url] = info
                    else:
                        results[url] = None
                else:
                    results[url] = None

        # Create response using Pydantic model
        response_data = ImportFailInfoBulkResponse(root=results)
        return web.json_response(response_data.root)
    except ValidationError as e:
        logging.error(f"[ComfyUI-Manager] Invalid request data: {e}")
        return web.Response(status=400, text=f"Invalid request data: {e}")
    except Exception as e:
        logging.error(f"[ComfyUI-Manager] Error processing bulk import fail info: {e}")
        return web.Response(status=500, text="Internal server error")


@routes.get("/v2/manager/queue/reset")
async def reset_queue(request):
    logging.debug("[ComfyUI-Manager] Queue reset requested")
    task_queue.wipe_queue()
    return web.Response(status=200)


@routes.get("/v2/manager/queue/status")
async def queue_count(request):
    """Get current queue status with optional client filtering.

    Query parameters:
        client_id: Optional client ID to filter tasks

    Returns:
        JSON with queue counts and processing status
    """
    client_id = request.query.get("client_id")

    if client_id:
        # Filter tasks by client_id
        running_client_tasks = [
            task
            for task in task_queue.running_tasks.values()
            if task.client_id == client_id
        ]
        pending_client_tasks = [
            item
            for priority, counter, item in task_queue.pending_tasks
            if item.client_id == client_id
        ]
        history_client_tasks = {
            ui_id: task
            for ui_id, task in task_queue.history_tasks.items()
            if hasattr(task, "client_id") and task.client_id == client_id
        }

        return web.json_response(
            {
                "client_id": client_id,
                "total_count": len(pending_client_tasks) + len(running_client_tasks),
                "done_count": len(history_client_tasks),
                "in_progress_count": len(running_client_tasks),
                "pending_count": len(pending_client_tasks),
                "is_processing": len(running_client_tasks) > 0,
            }
        )
    else:
        # Return overall status
        return web.json_response(
            {
                "total_count": task_queue.total_count(),
                "done_count": task_queue.done_count(),
                "in_progress_count": len(task_queue.running_tasks),
                "pending_count": len(task_queue.pending_tasks),
                "is_processing": task_queue.is_processing(),
            }
        )


@routes.get("/v2/manager/queue/start")
async def queue_start(request):
    logging.debug("[ComfyUI-Manager] Queue start requested")
    started = task_queue.start_worker()

    if started:
        logging.debug("[ComfyUI-Manager] Queue worker started successfully")
        return web.Response(status=200)  # Started successfully
    else:
        logging.debug("[ComfyUI-Manager] Queue worker already in progress")
        return web.Response(status=201)  # Already in-progress


@routes.get("/v2/manager/queue/update_comfyui")
async def update_comfyui(request):
    """Queue a ComfyUI update based on the configured update policy."""
    try:
        # Validate query parameters using Pydantic model
        query_params = UpdateComfyUIQueryParams.model_validate(
            dict(request.rel_url.query)
        )

        # Check if stable parameter was provided, otherwise use config
        if query_params.stable is None:
            is_stable = core.get_config()["update_policy"] != "nightly-comfyui"
        else:
            is_stable = query_params.stable

        client_id = query_params.client_id
        ui_id = query_params.ui_id
    except ValidationError as e:
        return web.json_response(
            {"error": "Validation error", "details": e.errors()}, status=400
        )

    # Create update-comfyui task
    task = QueueTaskItem(
        ui_id=ui_id,
        client_id=client_id,
        kind=OperationType.update_comfyui.value,
        params=UpdateComfyUIParams(is_stable=is_stable),
    )

    task_queue.put(task)
    return web.Response(status=200)


@routes.get("/v2/comfyui_manager/comfyui_versions")
async def comfyui_versions(request):
    try:
        res, current, latest = core.get_comfyui_versions()
        return web.json_response(
            {"versions": res, "current": current},
            status=200,
            content_type="application/json",
        )
    except Exception as e:
        logging.error(f"ComfyUI update fail: {e}")

    return web.Response(status=400)


@routes.get("/v2/comfyui_manager/comfyui_switch_version")
async def comfyui_switch_version(request):
    try:
        # Validate query parameters using Pydantic model
        query_params = ComfyUISwitchVersionQueryParams.model_validate(
            dict(request.rel_url.query)
        )

        target_version = query_params.ver
        client_id = query_params.client_id
        ui_id = query_params.ui_id

        # Create update-comfyui task with target version
        task = QueueTaskItem(
            ui_id=ui_id,
            client_id=client_id,
            kind=OperationType.update_comfyui.value,
            params=UpdateComfyUIParams(target_version=target_version),
        )

        task_queue.put(task)
        return web.Response(status=200)
    except ValidationError as e:
        return web.json_response(
            {"error": "Validation error", "details": e.errors()}, status=400
        )
    except Exception as e:
        logging.error(f"ComfyUI version switch fail: {e}")
        return web.Response(status=400)


@routes.post("/v2/manager/queue/install_model")
async def install_model(request):
    try:
        json_data = await request.json()

        # Validate required fields
        if "client_id" not in json_data:
            return web.Response(status=400, text="Missing required field: client_id")
        if "ui_id" not in json_data:
            return web.Response(status=400, text="Missing required field: ui_id")

        # Validate model metadata
        model_data = ModelMetadata.model_validate(json_data)

        # Create install-model task with client-provided IDs
        task = QueueTaskItem(
            ui_id=json_data["ui_id"],
            client_id=json_data["client_id"],
            kind=OperationType.install_model.value,
            params=model_data,
        )

        task_queue.put(task)
        return web.Response(status=200)
    except ValidationError as e:
        logging.error(f"[ComfyUI-Manager] Invalid model data: {e}")
        return web.Response(status=400, text=f"Invalid model data: {e}")
    except Exception as e:
        logging.error(f"[ComfyUI-Manager] Error processing model install: {e}")
        return web.Response(status=500, text="Internal server error")


@routes.get("/v2/manager/db_mode")
async def db_mode(request):
    if "value" in request.rel_url.query:
        environment_utils.set_db_mode(request.rel_url.query["value"])
        core.write_config()
    else:
        return web.Response(text=core.get_config()["db_mode"], status=200)

    return web.Response(status=200)


@routes.get("/v2/manager/policy/update")
async def update_policy(request):
    if "value" in request.rel_url.query:
        environment_utils.set_update_policy(request.rel_url.query["value"])
        core.write_config()
    else:
        return web.Response(text=core.get_config()["update_policy"], status=200)

    return web.Response(status=200)


@routes.get("/v2/manager/channel_url_list")
async def channel_url_list(request):
    channels = core.get_channel_dict()
    if "value" in request.rel_url.query:
        channel_url = channels.get(request.rel_url.query["value"])
        if channel_url is not None:
            core.get_config()["channel_url"] = channel_url
            core.write_config()
    else:
        selected = "custom"
        selected_url = core.get_config()["channel_url"]

        for name, url in channels.items():
            if url == selected_url:
                selected = name
                break

        res = {"selected": selected, "list": core.get_channel_list()}
        return web.json_response(res, status=200)

    return web.Response(status=200)


@routes.get("/v2/manager/reboot")
def restart(self):
    if not security_utils.is_allowed_security_level("middle"):
        logging.error(SECURITY_MESSAGE_MIDDLE)
        return web.Response(status=403)

    try:
        sys.stdout.close_log()
    except Exception:
        pass

    if "__COMFY_CLI_SESSION__" in os.environ:
        with open(os.path.join(os.environ["__COMFY_CLI_SESSION__"] + ".reboot"), "w"):
            pass

        print(
            "\nRestarting...\n\n"
        )  # This printing should not be logging - that will be ugly
        exit(0)

    print(
        "\nRestarting... [Legacy Mode]\n\n"
    )  # This printing should not be logging - that will be ugly

    sys_argv = sys.argv.copy()
    if "--windows-standalone-build" in sys_argv:
        sys_argv.remove("--windows-standalone-build")

    if sys_argv[0].endswith("__main__.py"):  # this is a python module
        module_name = os.path.basename(os.path.dirname(sys_argv[0]))
        cmds = [sys.executable, "-m", module_name] + sys_argv[1:]
    elif sys.platform.startswith("win32"):
        cmds = ['"' + sys.executable + '"', '"' + sys_argv[0] + '"'] + sys_argv[1:]
    else:
        cmds = [sys.executable] + sys_argv

    print(f"Command: {cmds}", flush=True)

    return os.execv(sys.executable, cmds)


@routes.get("/v2/manager/version")
async def get_version(request):
    return web.Response(text=core.version_str, status=200)


async def _confirm_try_install(sender, custom_node_url, msg):
    json_obj = await core.get_data_by_mode("default", "custom-node-list.json")

    sender = manager_util.sanitize_tag(sender)
    msg = manager_util.sanitize_tag(msg)
    target = core.lookup_customnode_by_url(json_obj, custom_node_url)

    if target is not None:
        PromptServer.instance.send_sync(
            "cm-api-try-install-customnode",
            {"sender": sender, "target": target, "msg": msg},
        )
    else:
        logging.error(
            f"[ComfyUI Manager API] Failed to try install - Unknown custom node url '{custom_node_url}'"
        )


def confirm_try_install(sender, custom_node_url, msg):
    asyncio.run(_confirm_try_install(sender, custom_node_url, msg))


cm_global.register_api("cm.try-install-custom-node", confirm_try_install)


async def default_cache_update():
    core.refresh_channel_dict()
    channel_url = core.get_config()["channel_url"]

    async def get_cache(filename):
        try:
            if core.get_config()["default_cache_as_channel_url"]:
                uri = f"{channel_url}/{filename}"
            else:
                uri = f"{core.DEFAULT_CHANNEL}/{filename}"

            cache_uri = str(manager_util.simple_hash(uri)) + "_" + filename
            cache_uri = os.path.join(manager_util.cache_dir, cache_uri)

            json_obj = await manager_util.get_data(uri, True)

            with manager_util.cache_lock:
                with open(cache_uri, "w", encoding="utf-8") as file:
                    json.dump(json_obj, file, indent=4, sort_keys=True)
                    logging.debug(f"[ComfyUI-Manager] default cache updated: {uri}")
        except Exception as e:
            logging.error(
                f"[ComfyUI-Manager] Failed to perform initial fetching '{filename}': {e}"
            )
            traceback.print_exc()

    if (
        core.get_config()["network_mode"] != "offline"
        and not manager_util.is_manager_pip_package()
    ):
        a = get_cache("custom-node-list.json")
        b = get_cache("extension-node-map.json")
        c = get_cache("model-list.json")
        d = get_cache("alter-list.json")
        e = get_cache("github-stats.json")

        await asyncio.gather(a, b, c, d, e)

        if core.get_config()["network_mode"] == "private":
            logging.info(
                "[ComfyUI-Manager] The private comfyregistry is not yet supported in `network_mode=private`."
            )
        else:
            # load at least once
            await core.unified_manager.reload(
                ManagerDatabaseSource.remote.value, dont_wait=False
            )
            await core.unified_manager.get_custom_nodes(
                channel_url, ManagerDatabaseSource.remote.value
            )
    else:
        await core.unified_manager.reload(
            ManagerDatabaseSource.remote.value, dont_wait=False, update_cnr_map=False
        )

    logging.info("[ComfyUI-Manager] All startup tasks have been completed.")


threading.Thread(target=lambda: asyncio.run(default_cache_update())).start()

if not os.path.exists(context.manager_config_path):
    core.get_config()
    core.write_config()




import os
import sys
import traceback
import json
import asyncio
import concurrent
import threading
from typing import Optional

import typer
from rich import print
from typing_extensions import List, Annotated
import re
import git
import importlib


from ..common import manager_util

# read env vars
# COMFYUI_FOLDERS_BASE_PATH is not required in cm-cli.py
# `comfy_path` should be resolved before importing manager_core

comfy_path = os.environ.get('COMFYUI_PATH')

if comfy_path is None:
    print("[bold red]cm-cli: environment variable 'COMFYUI_PATH' is not specified.[/bold red]")
    exit(-1)

sys.path.append(comfy_path)

if not os.path.exists(os.path.join(comfy_path, 'folder_paths.py')):
    print("[bold red]cm-cli: '{comfy_path}' is not a valid 'COMFYUI_PATH' location.[/bold red]")
    exit(-1)


import utils.extra_config
from ..common import cm_global
from ..legacy import manager_core as core
from ..common import context
from ..legacy.manager_core import unified_manager
from ..common import cnr_utils

comfyui_manager_path = os.path.abspath(os.path.dirname(__file__))

cm_global.pip_blacklist = {'torch', 'torchaudio', 'torchsde', 'torchvision'}
cm_global.pip_downgrade_blacklist = ['torch', 'torchaudio', 'torchsde', 'torchvision', 'transformers', 'safetensors', 'kornia']

cm_global.pip_overrides = {}

if os.path.exists(os.path.join(manager_util.comfyui_manager_path, "pip_overrides.json")):
    with open(os.path.join(manager_util.comfyui_manager_path, "pip_overrides.json"), 'r', encoding="UTF-8", errors="ignore") as json_file:
        cm_global.pip_overrides = json.load(json_file)


if os.path.exists(os.path.join(manager_util.comfyui_manager_path, "pip_blacklist.list")):
    with open(os.path.join(manager_util.comfyui_manager_path, "pip_blacklist.list"), 'r', encoding="UTF-8", errors="ignore") as f:
        for x in f.readlines():
            y = x.strip()
            if y != '':
                cm_global.pip_blacklist.add(y)


def check_comfyui_hash():
    try:
        repo = git.Repo(comfy_path)
        core.comfy_ui_revision = len(list(repo.iter_commits('HEAD')))
        core.comfy_ui_commit_datetime = repo.head.commit.committed_datetime
    except Exception:
        print('[bold yellow]INFO: Frozen ComfyUI mode.[/bold yellow]')
        core.comfy_ui_revision = 0
        core.comfy_ui_commit_datetime = 0

    cm_global.variables['comfyui.revision'] = core.comfy_ui_revision


check_comfyui_hash()  # This is a preparation step for manager_core
core.check_invalid_nodes()


def read_downgrade_blacklist():
    try:
        import configparser
        config = configparser.ConfigParser(strict=False)
        config.read(context.manager_config_path)
        default_conf = config['default']

        if 'downgrade_blacklist' in default_conf:
            items = default_conf['downgrade_blacklist'].split(',')
            items = [x.strip() for x in items if x != '']
            cm_global.pip_downgrade_blacklist += items
            cm_global.pip_downgrade_blacklist = list(set(cm_global.pip_downgrade_blacklist))
    except Exception:
        pass


read_downgrade_blacklist()  # This is a preparation step for manager_core


class Ctx:
    folder_paths = None
    
    def __init__(self):
        self.channel = 'default'
        self.no_deps = False
        self.mode = 'cache'
        self.user_directory = None
        self.custom_nodes_paths = [os.path.join(context.comfy_base_path, 'custom_nodes')]
        self.manager_files_directory = os.path.dirname(__file__)
        
        if Ctx.folder_paths is None:
            try:
                Ctx.folder_paths = importlib.import_module('folder_paths')
            except ImportError:
                print("Warning: Unable to import folder_paths module")

    def set_channel_mode(self, channel, mode):
        if mode is not None:
            self.mode = mode

        valid_modes = ["remote", "local", "cache"]
        if mode and mode.lower() not in valid_modes:
            typer.echo(
                f"Invalid mode: {mode}. Allowed modes are 'remote', 'local', 'cache'.",
                err=True,
            )
            exit(1)

        if channel is not None:
            self.channel = channel

        asyncio.run(unified_manager.reload(cache_mode=self.mode, dont_wait=False))
        asyncio.run(unified_manager.load_nightly(self.channel, self.mode))

    def set_no_deps(self, no_deps):
        self.no_deps = no_deps

    def set_user_directory(self, user_directory):
        if user_directory is None:
            return

        extra_model_paths_yaml = os.path.join(user_directory, 'extra_model_paths.yaml')
        if os.path.exists(extra_model_paths_yaml):
            utils.extra_config.load_extra_path_config(extra_model_paths_yaml)

        context.update_user_directory(user_directory)

        if os.path.exists(context.manager_pip_overrides_path):
            with open(context.manager_pip_overrides_path, 'r', encoding="UTF-8", errors="ignore") as json_file:
                cm_global.pip_overrides = json.load(json_file)

        if os.path.exists(context.manager_pip_blacklist_path):
            with open(context.manager_pip_blacklist_path, 'r', encoding="UTF-8", errors="ignore") as f:
                for x in f.readlines():
                    y = x.strip()
                    if y != '':
                        cm_global.pip_blacklist.add(y)

    def update_custom_nodes_dir(self, target_dir):
        import folder_paths
        a, b = folder_paths.folder_names_and_paths['custom_nodes']
        folder_paths.folder_names_and_paths['custom_nodes'] = [os.path.abspath(target_dir)], set()

    @staticmethod
    def get_startup_scripts_path():
        return os.path.join(context.manager_startup_script_path, "install-scripts.txt")

    @staticmethod
    def get_restore_snapshot_path():
        return os.path.join(context.manager_startup_script_path, "restore-snapshot.json")

    @staticmethod
    def get_snapshot_path():
        return context.manager_snapshot_path

    @staticmethod
    def get_custom_nodes_paths():
        if Ctx.folder_paths is None:
            print("Error: folder_paths module is not available")
            return []
        return Ctx.folder_paths.get_folder_paths('custom_nodes')


cmd_ctx = Ctx()


def install_node(node_spec_str, is_all=False, cnt_msg='', **kwargs):
    exit_on_fail = kwargs.get('exit_on_fail', False)
    print(f"install_node exit on fail:{exit_on_fail}...")
    
    if core.is_valid_url(node_spec_str):
        # install via urls
        res = asyncio.run(core.gitclone_install(node_spec_str, no_deps=cmd_ctx.no_deps))
        if not res.result:
            print(res.msg)
            print(f"[bold red]ERROR: An error occurred while installing '{node_spec_str}'.[/bold red]")
            if exit_on_fail:
                sys.exit(1)
        else:
            print(f"{cnt_msg} [INSTALLED] {node_spec_str:50}")
    else:
        node_spec = unified_manager.resolve_node_spec(node_spec_str)

        if node_spec is None:
            return

        node_name, version_spec, is_specified = node_spec

        # NOTE: install node doesn't allow update if version is not specified
        if not is_specified:
            version_spec = None

        res = asyncio.run(unified_manager.install_by_id(node_name, version_spec, cmd_ctx.channel, cmd_ctx.mode, instant_execution=True, no_deps=cmd_ctx.no_deps))

        if res.action == 'skip':
            print(f"{cnt_msg} [   SKIP  ] {node_name:50} => Already installed")
        elif res.action == 'enable':
            print(f"{cnt_msg} [ ENABLED ] {node_name:50}")
        elif res.action == 'install-git' and res.target == 'nightly':
            print(f"{cnt_msg} [INSTALLED] {node_name:50}[NIGHTLY]")
        elif res.action == 'install-git' and res.target == 'unknown':
            print(f"{cnt_msg} [INSTALLED] {node_name:50}[UNKNOWN]")
        elif res.action == 'install-cnr' and res.result:
            print(f"{cnt_msg} [INSTALLED] {node_name:50}[{res.target}]")
        elif res.action == 'switch-cnr' and res.result:
            print(f"{cnt_msg} [INSTALLED] {node_name:50}[{res.target}]")
        elif (res.action == 'switch-cnr' or res.action == 'install-cnr') and not res.result and node_name in unified_manager.cnr_map:
            print(f"\nAvailable version of '{node_name}'")
            show_versions(node_name)
            print("")
        else:
            print(f"[bold red]ERROR: An error occurred while installing '{node_name}'.\n{res.msg}[/bold red]")
            if exit_on_fail:
                sys.exit(1)


def reinstall_node(node_spec_str, is_all=False, cnt_msg=''):
    node_spec = unified_manager.resolve_node_spec(node_spec_str)

    node_name, version_spec, _ = node_spec

    unified_manager.unified_uninstall(node_name, version_spec == 'unknown')
    install_node(node_name, is_all=is_all, cnt_msg=cnt_msg)


def fix_node(node_spec_str, is_all=False, cnt_msg=''):
    node_spec = unified_manager.resolve_node_spec(node_spec_str, guess_mode='active')

    if node_spec is None:
        if not is_all:
            if unified_manager.resolve_node_spec(node_spec_str, guess_mode='inactive') is not None:
                print(f"{cnt_msg} [  SKIPPED  ]: {node_spec_str:50} => Disabled")
            else:
                print(f"{cnt_msg} [  SKIPPED  ]: {node_spec_str:50} => Not installed")

        return

    node_name, version_spec, _ = node_spec

    print(f"{cnt_msg} [   FIXING  ]: {node_name:50}[{version_spec}]")
    res = unified_manager.unified_fix(node_name, version_spec, no_deps=cmd_ctx.no_deps)

    if not res.result:
        print(f"[bold red]ERROR: f{res.msg}[/bold red]")


def uninstall_node(node_spec_str: str, is_all: bool = False, cnt_msg: str = ''):
    spec = node_spec_str.split('@')
    if len(spec) == 2 and spec[1] == 'unknown':
        node_name = spec[0]
        is_unknown = True
    else:
        node_name = spec[0]
        is_unknown = False

    res = unified_manager.unified_uninstall(node_name, is_unknown)
    if len(spec) == 1 and res.action == 'skip' and not is_unknown:
        res = unified_manager.unified_uninstall(node_name, True)

    if res.action == 'skip':
        print(f"{cnt_msg} [  SKIPPED  ]: {node_name:50} => Not installed")

    elif res.result:
        print(f"{cnt_msg} [UNINSTALLED] {node_name:50}")
    else:
        print(f"ERROR: An error occurred while uninstalling '{node_name}'.")


def update_node(node_spec_str, is_all=False, cnt_msg=''):
    node_spec = unified_manager.resolve_node_spec(node_spec_str, 'active')

    if node_spec is None:
        if unified_manager.resolve_node_spec(node_spec_str, 'inactive'):
            print(f"{cnt_msg} [  SKIPPED  ]: {node_spec_str:50} => Disabled")
        else:
            print(f"{cnt_msg} [  SKIPPED  ]: {node_spec_str:50} => Not installed")
        return None

    node_name, version_spec, _ = node_spec

    res = unified_manager.unified_update(node_name, version_spec, no_deps=cmd_ctx.no_deps, return_postinstall=True)

    if not res.result:
        print(f"ERROR: An error occurred while updating '{node_name}'.")
    elif res.action == 'skip':
        print(f"{cnt_msg} [  SKIPPED  ]: {node_name:50} => {res.msg}")
    else:
        print(f"{cnt_msg} [  UPDATED  ]: {node_name:50} => ({version_spec} -> {res.target})")

    return res.with_target(f'{node_name}@{res.target}')


def update_parallel(nodes):
    is_all = False
    if 'all' in nodes:
        is_all = True
        nodes = []
        for x in unified_manager.active_nodes.keys():
            nodes.append(x)
        for x in unified_manager.unknown_active_nodes.keys():
            nodes.append(x+"@unknown")
    else:
        nodes = [x for x in nodes if x.lower() not in ['comfy', 'comfyui']]

    total = len(nodes)

    lock = threading.Lock()
    processed = []

    i = 0

    def process_custom_node(x):
        nonlocal i
        nonlocal processed

        with lock:
            i += 1

        try:
            res = update_node(x, is_all=is_all, cnt_msg=f'{i}/{total}')
            with lock:
                processed.append(res)
        except Exception as e:
            print(f"ERROR: {e}")
            traceback.print_exc()

    with concurrent.futures.ThreadPoolExecutor(4) as executor:
        for item in nodes:
            executor.submit(process_custom_node, item)

    i = 1
    for res in processed:
        if res is not None:
            print(f"[{i}/{total}] Post update: {res.target}")
            if res.postinstall is not None:
                res.postinstall()
        i += 1


def update_comfyui():
    res = core.update_path(comfy_path, instant_execution=True)
    if res == 'fail':
        print("Updating ComfyUI has failed.")
    elif res == 'updated':
        print("ComfyUI is updated.")
    else:
        print("ComfyUI is already up to date.")


def enable_node(node_spec_str, is_all=False, cnt_msg=''):
    if unified_manager.resolve_node_spec(node_spec_str, guess_mode='active') is not None:
        print(f"{cnt_msg} [  SKIP ] {node_spec_str:50} => Already enabled")
        return

    node_spec = unified_manager.resolve_node_spec(node_spec_str, guess_mode='inactive')

    if node_spec is None:
        print(f"{cnt_msg} [  SKIP ] {node_spec_str:50} => Not found")
        return

    node_name, version_spec, _ = node_spec

    res = unified_manager.unified_enable(node_name, version_spec)

    if res.action == 'skip':
        print(f"{cnt_msg} [  SKIP ] {node_name:50} => {res.msg}")
    elif res.result:
        print(f"{cnt_msg} [ENABLED] {node_name:50}")
    else:
        print(f"{cnt_msg} [  FAIL ] {node_name:50} => {res.msg}")


def disable_node(node_spec_str: str, is_all=False, cnt_msg=''):
    if 'comfyui-manager' in node_spec_str.lower():
        return

    node_spec = unified_manager.resolve_node_spec(node_spec_str, guess_mode='active')

    if node_spec is None:
        if unified_manager.resolve_node_spec(node_spec_str, guess_mode='inactive') is not None:
            print(f"{cnt_msg} [  SKIP  ] {node_spec_str:50} => Already disabled")
        else:
            print(f"{cnt_msg} [  SKIP  ] {node_spec_str:50} => Not found")
        return

    node_name, version_spec, _ = node_spec

    res = unified_manager.unified_disable(node_name, version_spec == 'unknown')

    if res.action == 'skip':
        print(f"{cnt_msg} [  SKIP  ] {node_name:50} => {res.msg}")
    elif res.result:
        print(f"{cnt_msg} [DISABLED] {node_name:50}")
    else:
        print(f"{cnt_msg} [  FAIL  ] {node_name:50} => {res.msg}")


def show_list(kind, simple=False):
    custom_nodes = asyncio.run(unified_manager.get_custom_nodes(channel=cmd_ctx.channel, mode=cmd_ctx.mode))

    # collect not-installed unknown nodes
    not_installed_unknown_nodes = []
    repo_unknown = {}

    for k, v in custom_nodes.items():
        if 'cnr_latest' not in v:
            if len(v['files']) == 1:
                repo_url = v['files'][0]
                node_name = repo_url.split('/')[-1]
                if node_name not in unified_manager.unknown_inactive_nodes and node_name not in unified_manager.unknown_active_nodes:
                    not_installed_unknown_nodes.append(v)
                else:
                    repo_unknown[node_name] = v

    processed = {}
    unknown_processed = []

    flag = kind in ['all', 'cnr', 'installed', 'enabled']
    for k, v in unified_manager.active_nodes.items():
        if flag:
            cnr = unified_manager.cnr_map.get(k)
            if cnr:
                processed[k] = "[    ENABLED    ] ", cnr['name'], k, cnr['publisher']['name'], v[0]
            else:
                processed[k] = None
        else:
            processed[k] = None

    if flag and kind != 'cnr':
        for k, v in unified_manager.unknown_active_nodes.items():
            item = repo_unknown.get(k)

            if item is None:
                continue

            log_item = "[    ENABLED    ] ", item['title'], k, item['author']
            unknown_processed.append(log_item)

    flag = kind in ['all', 'cnr', 'installed', 'disabled']
    for k, v in unified_manager.cnr_inactive_nodes.items():
        if k in processed:
            continue

        if flag:
            cnr = unified_manager.cnr_map.get(k)  # NOTE: can this be None if removed from CNR after installed
            if cnr:
                processed[k] = "[    DISABLED   ] ", cnr['name'], k, cnr['publisher']['name'], ", ".join(list(v.keys()))
            else:
                processed[k] = None
        else:
            processed[k] = None

    for k, v in unified_manager.nightly_inactive_nodes.items():
        if k in processed:
            continue

        if flag:
            cnr = unified_manager.cnr_map.get(k)
            if cnr:
                processed[k] = "[    DISABLED   ] ", cnr['name'], k, cnr['publisher']['name'], 'nightly'
            else:
                processed[k] = None
        else:
            processed[k] = None

    if flag and kind != 'cnr':
        for k, v in unified_manager.unknown_inactive_nodes.items():
            item = repo_unknown.get(k)

            if item is None:
                continue

            log_item = "[    DISABLED   ] ", item['title'], k, item['author']
            unknown_processed.append(log_item)

    flag = kind in ['all', 'cnr', 'not-installed']
    for k, v in unified_manager.cnr_map.items():
        if k in processed:
            continue

        if flag:
            cnr = unified_manager.cnr_map.get(k)
            if cnr:
                ver_spec = v['latest_version']['version'] if 'latest_version' in v else '0.0.0'
                processed[k] = "[ NOT INSTALLED ] ", cnr['name'], k, cnr['publisher']['name'], ver_spec
            else:
                processed[k] = None
        else:
            processed[k] = None

    if flag and kind != 'cnr':
        for x in not_installed_unknown_nodes:
            if len(x['files']) == 1:
                node_id = os.path.basename(x['files'][0])
                log_item = "[ NOT INSTALLED ] ", x['title'], node_id, x['author']
                unknown_processed.append(log_item)

    for x in processed.values():
        if x is None:
            continue

        prefix, title, short_id, author, ver_spec = x
        if simple:
            print(title+'@'+ver_spec)
        else:
            print(f"{prefix} {title:50} {short_id:30} (author: {author:20}) \\[{ver_spec}]")

    for x in unknown_processed:
        prefix, title, short_id, author = x
        if simple:
            print(title+'@unknown')
        else:
            print(f"{prefix} {title:50} {short_id:30} (author: {author:20}) [UNKNOWN]")


async def show_snapshot(simple_mode=False):
    json_obj = await core.get_current_snapshot()

    if simple_mode:
        print(f"[{json_obj['comfyui']}] comfyui")
        for k, v in json_obj['git_custom_nodes'].items():
            print(f"[{v['hash']}] {k}")
        for v in json_obj['file_custom_nodes']:
            print(f"[                   N/A                  ] {v['filename']}")

    else:
        formatted_json = json.dumps(json_obj, ensure_ascii=False, indent=4)
        print(formatted_json)


def show_snapshot_list(simple_mode=False):
    snapshot_path = cmd_ctx.get_snapshot_path()

    files = os.listdir(snapshot_path)
    json_files = [x for x in files if x.endswith('.json')]
    for x in sorted(json_files):
        print(x)


def cancel():
    if os.path.exists(cmd_ctx.get_startup_scripts_path()):
        os.remove(cmd_ctx.get_startup_scripts_path())

    if os.path.exists(cmd_ctx.get_restore_snapshot_path()):
        os.remove(cmd_ctx.get_restore_snapshot_path())


async def auto_save_snapshot():
    path = await core.save_snapshot_with_postfix('cli-autosave')
    print(f"Current snapshot is saved as `{path}`")


def get_all_installed_node_specs():
    res = []
    processed = set()
    for k, v in unified_manager.active_nodes.items():
        node_spec_str = f"{k}@{v[0]}"
        res.append(node_spec_str)
        processed.add(k)

    for k in unified_manager.cnr_inactive_nodes.keys():
        if k in processed:
            continue

        latest = unified_manager.get_from_cnr_inactive_nodes(k)
        if latest is not None:
            node_spec_str = f"{k}@{str(latest[0])}"
            res.append(node_spec_str)

    for k in unified_manager.nightly_inactive_nodes.keys():
        if k in processed:
            continue

        node_spec_str = f"{k}@nightly"
        res.append(node_spec_str)

    for k in unified_manager.unknown_active_nodes.keys():
        node_spec_str = f"{k}@unknown"
        res.append(node_spec_str)

    for k in unified_manager.unknown_inactive_nodes.keys():
        node_spec_str = f"{k}@unknown"
        res.append(node_spec_str)

    return res


def for_each_nodes(nodes, act, allow_all=True, **kwargs):
    is_all = False
    if allow_all and 'all' in nodes:
        is_all = True
        nodes = get_all_installed_node_specs()
    else:
        nodes = [x for x in nodes if x.lower() not in ['comfy', 'comfyui', 'all']]

    total = len(nodes)
    i = 1
    for x in nodes:
        try:
            act(x, is_all=is_all, cnt_msg=f'{i}/{total}', **kwargs)
        except Exception as e:
            print(f"ERROR: {e}")
            traceback.print_exc()
        i += 1


app = typer.Typer()


@app.command(help="Display help for commands")
def help(ctx: typer.Context):
    print(ctx.find_root().get_help())
    ctx.exit(0)


@app.command(help="Install custom nodes")
def install(
        nodes: List[str] = typer.Argument(
            ..., help="List of custom nodes to install"
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        no_deps: Annotated[
            Optional[bool],
            typer.Option(
                "--no-deps",
                show_default=False,
                help="Skip installing any Python dependencies",
            ),
        ] = False,
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
        exit_on_fail: bool = typer.Option(
            False,
            help="Exit on failure"
        )
):
    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)
    cmd_ctx.set_no_deps(no_deps)

    pip_fixer = manager_util.PIPFixer(manager_util.get_installed_packages(), comfy_path, context.manager_files_path)
    for_each_nodes(nodes, act=install_node, exit_on_fail=exit_on_fail)
    pip_fixer.fix_broken()


@app.command(help="Reinstall custom nodes")
def reinstall(
        nodes: List[str] = typer.Argument(
            ..., help="List of custom nodes to reinstall"
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        no_deps: Annotated[
            Optional[bool],
            typer.Option(
                "--no-deps",
                show_default=False,
                help="Skip installing any Python dependencies",
            ),
        ] = False,
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
):
    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)
    cmd_ctx.set_no_deps(no_deps)

    pip_fixer = manager_util.PIPFixer(manager_util.get_installed_packages(), comfy_path, context.manager_files_path)
    for_each_nodes(nodes, act=reinstall_node)
    pip_fixer.fix_broken()


@app.command(help="Uninstall custom nodes")
def uninstall(
        nodes: List[str] = typer.Argument(
            ..., help="List of custom nodes to uninstall"
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
):
    cmd_ctx.set_channel_mode(channel, mode)
    for_each_nodes(nodes, act=uninstall_node)


@app.command(help="Update custom nodes")
def update(
        nodes: List[str] = typer.Argument(
            ...,
            help="[all|List of custom nodes to update]"
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
):
    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)

    if 'all' in nodes:
        asyncio.run(auto_save_snapshot())

    pip_fixer = manager_util.PIPFixer(manager_util.get_installed_packages(), comfy_path, context.manager_files_path)

    for x in nodes:
        if x.lower() in ['comfyui', 'comfy', 'all']:
            update_comfyui()
            break

    update_parallel(nodes)
    pip_fixer.fix_broken()


@app.command(help="Disable custom nodes")
def disable(
        nodes: List[str] = typer.Argument(
            ...,
            help="[all|List of custom nodes to disable]"
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
):
    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)

    if 'all' in nodes:
        asyncio.run(auto_save_snapshot())

    for_each_nodes(nodes, disable_node, allow_all=True)


@app.command(help="Enable custom nodes")
def enable(
        nodes: List[str] = typer.Argument(
            ...,
            help="[all|List of custom nodes to enable]"
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
):
    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)

    if 'all' in nodes:
        asyncio.run(auto_save_snapshot())

    for_each_nodes(nodes, enable_node, allow_all=True)


@app.command(help="Fix dependencies of custom nodes")
def fix(
        nodes: List[str] = typer.Argument(
            ...,
            help="[all|List of custom nodes to fix]"
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
):
    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)

    if 'all' in nodes:
        asyncio.run(auto_save_snapshot())

    pip_fixer = manager_util.PIPFixer(manager_util.get_installed_packages(), comfy_path, context.manager_files_path)
    for_each_nodes(nodes, fix_node, allow_all=True)
    pip_fixer.fix_broken()


@app.command("show-versions", help="Show all available versions of the node")
def show_versions(node_name: str):
    versions = cnr_utils.all_versions_of_node(node_name)
    if versions is None:
        print(f"Node not found in Comfy Registry: {node_name}")

    for x in versions:
        print(f"[{x['createdAt'][:10]}] {x['version']} -- {x['changelog']}")


@app.command("show", help="Show node list")
def show(
        arg: str = typer.Argument(
            help="[installed|enabled|not-installed|disabled|all|cnr|snapshot|snapshot-list]"
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
):
    valid_commands = [
        "installed",
        "enabled",
        "not-installed",
        "disabled",
        "all",
        "cnr",
        "snapshot",
        "snapshot-list",
    ]
    if arg not in valid_commands:
        typer.echo(f"Invalid command: `show {arg}`", err=True)
        exit(1)

    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)
    if arg == 'snapshot':
        show_snapshot()
    elif arg == 'snapshot-list':
        show_snapshot_list()
    else:
        show_list(arg)


@app.command("simple-show", help="Show node list (simple mode)")
def simple_show(
        arg: str = typer.Argument(
            help="[installed|enabled|not-installed|disabled|all|snapshot|snapshot-list]"
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
):
    valid_commands = [
        "installed",
        "enabled",
        "not-installed",
        "disabled",
        "all",
        "snapshot",
        "snapshot-list",
    ]
    if arg not in valid_commands:
        typer.echo(f"[bold red]Invalid command: `show {arg}`[/bold red]", err=True)
        exit(1)

    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)

    if arg == 'snapshot':
        show_snapshot(True)
    elif arg == 'snapshot-list':
        show_snapshot_list(True)
    else:
        show_list(arg, True)


@app.command('cli-only-mode', help="Set whether to use ComfyUI-Manager in CLI-only mode.")
def cli_only_mode(
        mode: str = typer.Argument(
            ..., help="[enable|disable]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        )
):
    cmd_ctx.set_user_directory(user_directory)
    cli_mode_flag = os.path.join(cmd_ctx.manager_files_directory, '.enable-cli-only-mode')

    if mode.lower() == 'enable':
        with open(cli_mode_flag, 'w'):
            pass
        print("\nINFO: `cli-only-mode` is enabled\n")
    elif mode.lower() == 'disable':
        if os.path.exists(cli_mode_flag):
            os.remove(cli_mode_flag)
        print("\nINFO: `cli-only-mode` is disabled\n")
    else:
        print(f"\n[bold red]Invalid value for cli-only-mode: {mode}[/bold red]\n")
        exit(1)


@app.command(
    "deps-in-workflow", help="Generate dependencies file from workflow (.json/.png)"
)
def deps_in_workflow(
        workflow: Annotated[
            str, typer.Option(show_default=False, help="Workflow file (.json/.png)")
        ],
        output: Annotated[
            str, typer.Option(show_default=False, help="Output file (.json)")
        ],
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        )
):
    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)

    input_path = workflow
    output_path = output

    if not os.path.exists(input_path):
        print(f"[bold red]File not found: {input_path}[/bold red]")
        exit(1)

    used_exts, unknown_nodes = asyncio.run(core.extract_nodes_from_workflow(input_path, mode=cmd_ctx.mode, channel_url=cmd_ctx.channel))

    custom_nodes = {}
    for x in used_exts:
        custom_nodes[x] = {'state': core.simple_check_custom_node(x),
                           'hash': '-'
                           }

    res = {
        'custom_nodes': custom_nodes,
        'unknown_nodes': list(unknown_nodes)
    }

    with open(output_path, "w", encoding='utf-8') as output_file:
        json.dump(res, output_file, indent=4)

    print(f"Workflow dependencies are being saved into {output_path}.")


@app.command("save-snapshot", help="Save a snapshot of the current ComfyUI environment. If output path isn't provided. Save to ComfyUI-Manager/snapshots path.")
def save_snapshot(
        output: Annotated[
            str,
            typer.Option(
                show_default=False, help="Specify the output file path. (.json/.yaml)"
            ),
        ] = None,
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
        full_snapshot: Annotated[
            bool,
            typer.Option(
                show_default=False, help="If the snapshot should include custom node, ComfyUI version and pip versions (default), or only custom node details"
            ),
        ] = True,
):
    cmd_ctx.set_user_directory(user_directory)

    if output is not None:
        if(not output.endswith('.json') and not output.endswith('.yaml')):
            print("[bold red]ERROR: output path should be either '.json' or '.yaml' file.[/bold red]")
            raise typer.Exit(code=1)
    
        dir_path = os.path.dirname(output)
        
        if(dir_path != '' and not os.path.exists(dir_path)):
            print(f"[bold red]ERROR: {output} path not exists.[/bold red]")
            raise typer.Exit(code=1)
        
    path = asyncio.run(core.save_snapshot_with_postfix('snapshot', output, not full_snapshot))
    print(f"Current snapshot is saved as `{path}`")


@app.command("restore-snapshot", help="Restore snapshot from snapshot file")
def restore_snapshot(
        snapshot_name: str, 
        pip_non_url: Optional[bool] = typer.Option(
            default=None,
            show_default=False,
            is_flag=True,
            help="Restore for pip packages registered on PyPI.",
        ),
        pip_non_local_url: Optional[bool] = typer.Option(
            default=None,
            show_default=False,
            is_flag=True,
            help="Restore for pip packages registered at web URLs.",
        ),
        pip_local_url: Optional[bool] = typer.Option(
            default=None,
            show_default=False,
            is_flag=True,
            help="Restore for pip packages specified by local paths.",
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
        restore_to: Optional[str] = typer.Option(
            None,
            help="Manually specify the installation path for the custom node. Ignore user directory."
        )
):
    cmd_ctx.set_user_directory(user_directory)

    if restore_to:
        cmd_ctx.update_custom_nodes_dir(restore_to)

    extras = []
    if pip_non_url:
        extras.append('--pip-non-url')

    if pip_non_local_url:
        extras.append('--pip-non-local-url')

    if pip_local_url:
        extras.append('--pip-local-url')

    print(f"PIPs restore mode: {extras}")

    if os.path.exists(snapshot_name):
        snapshot_path = os.path.abspath(snapshot_name)
    else:
        snapshot_path = os.path.join(cmd_ctx.get_snapshot_path(), snapshot_name)
        if not os.path.exists(snapshot_path):
            print(f"[bold red]ERROR: `{snapshot_path}` is not exists.[/bold red]")
            exit(1)

    pip_fixer = manager_util.PIPFixer(manager_util.get_installed_packages(), comfy_path, context.manager_files_path)
    try:
        asyncio.run(core.restore_snapshot(snapshot_path, extras))
    except Exception:
        print("[bold red]ERROR: Failed to restore snapshot.[/bold red]")
        traceback.print_exc()
        raise typer.Exit(code=1)
    pip_fixer.fix_broken()


@app.command(
    "restore-dependencies", help="Restore dependencies from whole installed custom nodes."
)
def restore_dependencies(
        user_directory: str = typer.Option(
            None,
            help="user directory"
        )
):
    cmd_ctx.set_user_directory(user_directory)

    node_paths = []

    for base_path in cmd_ctx.get_custom_nodes_paths():
        for name in os.listdir(base_path):
            target = os.path.join(base_path, name)
            if os.path.isdir(target) and not name.endswith('.disabled'):
                node_paths.append(target)

    total = len(node_paths)
    i = 1

    pip_fixer = manager_util.PIPFixer(manager_util.get_installed_packages(), comfy_path, context.manager_files_path)
    for x in node_paths:
        print("----------------------------------------------------------------------------------------------------")
        print(f"Restoring [{i}/{total}]: {x}")
        unified_manager.execute_install_script('', x, instant_execution=True)
        i += 1
    pip_fixer.fix_broken()


@app.command(
    "post-install", help="Install dependencies and execute installation script"
)
def post_install(
        path: str = typer.Argument(
            help="path to custom node",
        )
):
    path = os.path.expanduser(path)

    pip_fixer = manager_util.PIPFixer(manager_util.get_installed_packages(), comfy_path, context.manager_files_path)
    unified_manager.execute_install_script('', path, instant_execution=True)
    pip_fixer.fix_broken()


@app.command(
    "install-deps",
    help="Install dependencies from dependencies file(.json) or workflow(.png/.json)",
)
def install_deps(
        deps: str = typer.Argument(
            help="Dependency spec file (.json)",
        ),
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
):
    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)
    asyncio.run(auto_save_snapshot())

    if not os.path.exists(deps):
        print(f"[bold red]File not found: {deps}[/bold red]")
        exit(1)
    else:
        with open(deps, 'r', encoding="UTF-8", errors="ignore") as json_file:
            try:
                json_obj = json.load(json_file)
            except Exception:
                print(f"[bold red]Invalid json file: {deps}[/bold red]")
                exit(1)

            pip_fixer = manager_util.PIPFixer(manager_util.get_installed_packages(), comfy_path, context.manager_files_path)
            for k in json_obj['custom_nodes'].keys():
                state = core.simple_check_custom_node(k)
                if state == 'installed':
                    continue
                elif state == 'not-installed':
                    asyncio.run(core.gitclone_install(k, instant_execution=True))
                else:  # disabled
                    core.gitclone_set_active([k], False)
            pip_fixer.fix_broken()

        print("Dependency installation and activation complete.")


@app.command(help="Clear reserved startup action in ComfyUI-Manager")
def clear():
    cancel()


@app.command("export-custom-node-ids", help="Export custom node ids")
def export_custom_node_ids(
        path: str,
        channel: Annotated[
            str,
            typer.Option(
                show_default=False,
                help="Specify the operation mode"
            ),
        ] = None,
        mode: str = typer.Option(
            None,
            help="[remote|local|cache]"
        ),
        user_directory: str = typer.Option(
            None,
            help="user directory"
        ),
):
    cmd_ctx.set_user_directory(user_directory)
    cmd_ctx.set_channel_mode(channel, mode)

    with open(path, "w", encoding='utf-8') as output_file:
        for x in unified_manager.cnr_map.keys():
            print(x, file=output_file)

        custom_nodes = asyncio.run(unified_manager.get_custom_nodes(channel=cmd_ctx.channel, mode=cmd_ctx.mode))
        for x in custom_nodes.values():
            if 'cnr_latest' not in x:
                if len(x['files']) == 1:
                    repo_url = x['files'][0]
                    node_id = repo_url.split('/')[-1]
                    print(f"{node_id}@unknown", file=output_file)

                if 'id' in x:
                    print(f"{x['id']}@unknown", file=output_file)


def main():
    app()


if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
    sys.exit(app())


print("")



import os
import configparser


GITHUB_ENDPOINT = os.getenv('GITHUB_ENDPOINT')


def is_git_repo(path: str) -> bool:
    """ Check if the path is a git repository. """
    # NOTE: Checking it through `git.Repo` must be avoided.
    #       It locks the file, causing issues on Windows.
    return os.path.exists(os.path.join(path, '.git'))


def get_commit_hash(fullpath):
    git_head = os.path.join(fullpath, '.git', 'HEAD')
    if os.path.exists(git_head):
        with open(git_head) as f:
            line = f.readline()

            if line.startswith("ref: "):
                ref = os.path.join(fullpath, '.git', line[5:].strip())
                if os.path.exists(ref):
                    with open(ref) as f2:
                        return f2.readline().strip()
                else:
                    return "unknown"
            else:
                return line

    return "unknown"


def git_url(fullpath):
    """
    resolve version of unclassified custom node based on remote url in .git/config
    """
    git_config_path = os.path.join(fullpath, '.git', 'config')

    if not os.path.exists(git_config_path):
        return None

    # Set `strict=False` to allow duplicate `vscode-merge-base` sections, addressing <https://github.com/ltdrdata/ComfyUI-Manager/issues/1529>
    config = configparser.ConfigParser(strict=False)
    config.read(git_config_path)

    for k, v in config.items():
        if k.startswith('remote ') and 'url' in v:
            if 'Comfy-Org/ComfyUI-Manager' in v['url']:
                return "https://github.com/ltdrdata/ComfyUI-Manager"
            return v['url']

    return None


def normalize_url(url) -> str:
    github_id = normalize_to_github_id(url)
    if github_id is not None:
        url = f"https://github.com/{github_id}"

    return url


def normalize_to_github_id(url) -> str:
    if 'github' in url or (GITHUB_ENDPOINT is not None and GITHUB_ENDPOINT in url):
        author = os.path.basename(os.path.dirname(url))

        if author.startswith('git@github.com:'):
            author = author.split(':')[1]

        repo_name = os.path.basename(url)
        if repo_name.endswith('.git'):
            repo_name = repo_name[:-4]

        return f"{author}/{repo_name}"

    return None


def get_url_for_clone(url):
    url = normalize_url(url)

    if GITHUB_ENDPOINT is not None and url.startswith('https://github.com/'):
        url = GITHUB_ENDPOINT + url[18:] # url[18:] -> remove `https://github.com`

    return url
    


import subprocess
import sys
import os
import traceback

import git
import json
import yaml
import requests
from tqdm.auto import tqdm
from git.remote import RemoteProgress


comfy_path = os.environ.get('COMFYUI_PATH')
git_exe_path = os.environ.get('GIT_EXE_PATH')

if comfy_path is None:
    print("git_helper: environment variable 'COMFYUI_PATH' is not specified.")
    exit(-1)

if not os.path.exists(os.path.join(comfy_path, 'folder_paths.py')):
    print("git_helper: '{comfy_path}' is not a valid 'COMFYUI_PATH' location.")
    exit(-1)

def download_url(url, dest_folder, filename=None):
    # Ensure the destination folder exists
    if not os.path.exists(dest_folder):
        os.makedirs(dest_folder)

    # Extract filename from URL if not provided
    if filename is None:
        filename = os.path.basename(url)

    # Full path to save the file
    dest_path = os.path.join(dest_folder, filename)

    # Download the file
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        with open(dest_path, 'wb') as file:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    file.write(chunk)
    else:
        print(f"Failed to download file from {url}")


nodelist_path = os.path.join(os.path.dirname(__file__), "custom-node-list.json")
working_directory = os.getcwd()

if os.path.basename(working_directory) != 'custom_nodes':
    print("WARN: This script should be executed in custom_nodes dir")
    print(f"DBG: INFO {working_directory}")
    print(f"DBG: INFO {sys.argv}")
    # exit(-1)


class GitProgress(RemoteProgress):
    def __init__(self):
        super().__init__()
        self.pbar = tqdm(ascii=True)

    def update(self, op_code, cur_count, max_count=None, message=''):
        self.pbar.total = max_count
        self.pbar.n = cur_count
        self.pbar.pos = 0
        self.pbar.refresh()


def gitclone(custom_nodes_path, url, target_hash=None, repo_path=None):
    repo_name = os.path.splitext(os.path.basename(url))[0]

    if repo_path is None:
        repo_path = os.path.join(custom_nodes_path, repo_name)

    # Clone the repository from the remote URL
    repo = git.Repo.clone_from(url, repo_path, recursive=True, progress=GitProgress())

    if target_hash is not None:
        print(f"CHECKOUT: {repo_name} [{target_hash}]")
        repo.git.checkout(target_hash)
            
    repo.git.clear_cache()
    repo.close()


def gitcheck(path, do_fetch=False):
    try:
        # Fetch the latest commits from the remote repository
        repo = git.Repo(path)

        if repo.head.is_detached:
            print("CUSTOM NODE CHECK: True")
            return

        current_branch = repo.active_branch
        branch_name = current_branch.name

        remote_name = current_branch.tracking_branch().remote_name
        remote = repo.remote(name=remote_name)

        if do_fetch:
            remote.fetch()

        # Get the current commit hash and the commit hash of the remote branch
        commit_hash = repo.head.commit.hexsha

        if f'{remote_name}/{branch_name}' in repo.refs:
            remote_commit_hash = repo.refs[f'{remote_name}/{branch_name}'].object.hexsha
        else:
            print("CUSTOM NODE CHECK: True")  # non default branch is treated as updatable
            return

        # Compare the commit hashes to determine if the local repository is behind the remote repository
        if commit_hash != remote_commit_hash:
            # Get the commit dates
            commit_date = repo.head.commit.committed_datetime
            remote_commit_date = repo.refs[f'{remote_name}/{branch_name}'].object.committed_datetime

            # Compare the commit dates to determine if the local repository is behind the remote repository
            if commit_date < remote_commit_date:
                print("CUSTOM NODE CHECK: True")
        else:
            print("CUSTOM NODE CHECK: False")
    except Exception as e:
        print(e)
        print("CUSTOM NODE CHECK: Error")


def get_remote_name(repo):
    available_remotes = [remote.name for remote in repo.remotes]
    if 'origin' in available_remotes:
        return 'origin'
    elif 'upstream' in available_remotes:
        return 'upstream'
    elif len(available_remotes) > 0:
        return available_remotes[0]

    if not available_remotes:
        print(f"[ComfyUI-Manager] No remotes are configured for this repository: {repo.working_dir}")
    else:
        print(f"[ComfyUI-Manager] Available remotes in '{repo.working_dir}': ")
        for remote in available_remotes:
            print(f"- {remote}")

    return None


def switch_to_default_branch(repo):
    remote_name = get_remote_name(repo)

    try:
        if remote_name is None:
            return False

        default_branch = repo.git.symbolic_ref(f'refs/remotes/{remote_name}/HEAD').replace(f'refs/remotes/{remote_name}/', '')
        repo.git.checkout(default_branch)
        return True
    except Exception:
        # try checkout master
        # try checkout main if failed
        try:
            repo.git.checkout(repo.heads.master)
            return True
        except Exception:
            try:
                if remote_name is not None:
                    repo.git.checkout('-b', 'master', f'{remote_name}/master')
                    return True
            except Exception:
                try:
                    repo.git.checkout(repo.heads.main)
                    return True
                except Exception:
                    try:
                        if remote_name is not None:
                            repo.git.checkout('-b', 'main', f'{remote_name}/main')
                            return True
                    except Exception:
                        pass

    print("[ComfyUI Manager] Failed to switch to the default branch")
    return False


def gitpull(path):
    # Check if the path is a git repository
    if not os.path.exists(os.path.join(path, '.git')):
        raise ValueError('Not a git repository')

    # Pull the latest changes from the remote repository
    repo = git.Repo(path)
    if repo.is_dirty():
        print(f"STASH: '{path}' is dirty.")
        repo.git.stash()

    commit_hash = repo.head.commit.hexsha
    try:
        if repo.head.is_detached:
            switch_to_default_branch(repo)

        current_branch = repo.active_branch
        branch_name = current_branch.name

        remote_name = current_branch.tracking_branch().remote_name
        remote = repo.remote(name=remote_name)

        if f'{remote_name}/{branch_name}' not in repo.refs:
            switch_to_default_branch(repo)
            current_branch = repo.active_branch
            branch_name = current_branch.name

        remote.fetch()
        if f'{remote_name}/{branch_name}' in repo.refs:
            remote_commit_hash = repo.refs[f'{remote_name}/{branch_name}'].object.hexsha
        else:
            print("CUSTOM NODE PULL: Fail")  # update fail
            return

        if commit_hash == remote_commit_hash:
            print("CUSTOM NODE PULL: None")  # there is no update
            repo.close()
            return

        remote.pull()

        repo.git.submodule('update', '--init', '--recursive')
        new_commit_hash = repo.head.commit.hexsha

        if commit_hash != new_commit_hash:
            print("CUSTOM NODE PULL: Success")  # update success
        else:
            print("CUSTOM NODE PULL: Fail")  # update fail
    except Exception as e:
        print(e)
        print("CUSTOM NODE PULL: Fail")  # unknown git error

    repo.close()


def checkout_comfyui_hash(target_hash):
    repo = git.Repo(comfy_path)
    commit_hash = repo.head.commit.hexsha

    if commit_hash != target_hash:
        try:
            print(f"CHECKOUT: ComfyUI [{target_hash}]")
            repo.git.checkout(target_hash)
        except git.GitCommandError as e:
            print(f"Error checking out the ComfyUI: {str(e)}")


def checkout_custom_node_hash(git_custom_node_infos):
    repo_name_to_url = {}

    for url in git_custom_node_infos.keys():
        repo_name = url.split('/')[-1]

        if repo_name.endswith('.git'):
            repo_name = repo_name[:-4]

        repo_name_to_url[repo_name] = url

    for path in os.listdir(working_directory):
        if path.endswith("ComfyUI-Manager"):
            continue

        fullpath = os.path.join(working_directory, path)

        if os.path.isdir(fullpath):
            is_disabled = path.endswith(".disabled")

            try:
                git_dir = os.path.join(fullpath, '.git')
                if not os.path.exists(git_dir):
                    continue

                need_checkout = False
                repo_name = os.path.basename(fullpath)

                if repo_name.endswith('.disabled'):
                    repo_name = repo_name[:-9]

                if repo_name not in repo_name_to_url:
                    if not is_disabled:
                        # should be disabled
                        print(f"DISABLE: {repo_name}")
                        new_path = fullpath + ".disabled"
                        os.rename(fullpath, new_path)
                        need_checkout = False
                else:
                    item = git_custom_node_infos[repo_name_to_url[repo_name]]
                    if item['disabled'] and is_disabled:
                        pass
                    elif item['disabled'] and not is_disabled:
                        # disable
                        print(f"DISABLE: {repo_name}")
                        new_path = fullpath + ".disabled"
                        os.rename(fullpath, new_path)

                    elif not item['disabled'] and is_disabled:
                        # enable
                        print(f"ENABLE: {repo_name}")
                        new_path = fullpath[:-9]
                        os.rename(fullpath, new_path)
                        fullpath = new_path
                        need_checkout = True
                    else:
                        need_checkout = True

                if need_checkout:
                    repo = git.Repo(fullpath)
                    commit_hash = repo.head.commit.hexsha

                    if commit_hash != item['hash']:
                        print(f"CHECKOUT: {repo_name} [{item['hash']}]")
                        repo.git.checkout(item['hash'])

            except Exception:
                print(f"Failed to restore snapshots for the custom node '{path}'")

    # clone missing
    for k, v in git_custom_node_infos.items():
        if 'ComfyUI-Manager' in k:
            continue

        if not v['disabled']:
            repo_name = k.split('/')[-1]
            if repo_name.endswith('.git'):
                repo_name = repo_name[:-4]

            path = os.path.join(working_directory, repo_name)
            if not os.path.exists(path):
                print(f"CLONE: {path}")
                gitclone(working_directory, k, target_hash=v['hash'])


def invalidate_custom_node_file(file_custom_node_infos):
    global nodelist_path

    enabled_set = set()
    for item in file_custom_node_infos:
        if not item['disabled']:
            enabled_set.add(item['filename'])

    for path in os.listdir(working_directory):
        fullpath = os.path.join(working_directory, path)

        if not os.path.isdir(fullpath) and fullpath.endswith('.py'):
            if path not in enabled_set:
                print(f"DISABLE: {path}")
                new_path = fullpath+'.disabled'
                os.rename(fullpath, new_path)

        elif not os.path.isdir(fullpath) and fullpath.endswith('.py.disabled'):
            path = path[:-9]
            if path in enabled_set:
                print(f"ENABLE: {path}")
                new_path = fullpath[:-9]
                os.rename(fullpath, new_path)

    # download missing: just support for 'copy' style
    py_to_url = {}

    with open(nodelist_path, 'r', encoding="UTF-8") as json_file:
        info = json.load(json_file)
        for item in info['custom_nodes']:
            if item['install_type'] == 'copy':
                for url in item['files']:
                    if url.endswith('.py'):
                        py = url.split('/')[-1]
                        py_to_url[py] = url

        for item in file_custom_node_infos:
            filename = item['filename']
            if not item['disabled']:
                target_path = os.path.join(working_directory, filename)

                if not os.path.exists(target_path) and filename in py_to_url:
                    url = py_to_url[filename]
                    print(f"DOWNLOAD: {filename}")
                    download_url(url, working_directory)


def apply_snapshot(path):
    try:
        if os.path.exists(path):
            if not path.endswith('.json') and not path.endswith('.yaml'):
                print(f"Snapshot file not found: `{path}`")
                print("APPLY SNAPSHOT: False")
                return None

            with open(path, 'r', encoding="UTF-8") as snapshot_file:
                if path.endswith('.json'):
                    info = json.load(snapshot_file)
                elif path.endswith('.yaml'):
                    info = yaml.load(snapshot_file, Loader=yaml.SafeLoader)
                    info = info['custom_nodes']
                else:
                    # impossible case
                    print("APPLY SNAPSHOT: False")
                    return None

                comfyui_hash = info['comfyui']
                git_custom_node_infos = info['git_custom_nodes']
                file_custom_node_infos = info['file_custom_nodes']

                if comfyui_hash:
                    checkout_comfyui_hash(comfyui_hash)
                checkout_custom_node_hash(git_custom_node_infos)
                invalidate_custom_node_file(file_custom_node_infos)

                print("APPLY SNAPSHOT: True")
                if 'pips' in info and info['pips']:
                    return info['pips']
                else:
                    return None

        print(f"Snapshot file not found: `{path}`")
        print("APPLY SNAPSHOT: False")

        return None
    except Exception as e:
        print(e)
        traceback.print_exc()
        print("APPLY SNAPSHOT: False")

        return None


def restore_pip_snapshot(pips, options):
    non_url = []
    local_url = []
    non_local_url = []
    for k, v in pips.items():
        if v == "":
            non_url.append(k)
        else:
            if v.startswith('file:'):
                local_url.append(v)
            else:
                non_local_url.append(v)

    failed = []
    if '--pip-non-url' in options:
        # try all at once
        res = 1
        try:
            res = subprocess.check_call([sys.executable, '-m', 'pip', 'install'] + non_url)
        except Exception:
            pass

        # fallback
        if res != 0:
            for x in non_url:
                res = 1
                try:
                    res = subprocess.check_call([sys.executable, '-m', 'pip', 'install', x])
                except Exception:
                    pass

                if res != 0:
                    failed.append(x)

    if '--pip-non-local-url' in options:
        for x in non_local_url:
            res = 1
            try:
                res = subprocess.check_call([sys.executable, '-m', 'pip', 'install', x])
            except Exception:
                pass

            if res != 0:
                failed.append(x)

    if '--pip-local-url' in options:
        for x in local_url:
            res = 1
            try:
                res = subprocess.check_call([sys.executable, '-m', 'pip', 'install', x])
            except Exception:
                pass

            if res != 0:
                failed.append(x)

    print(f"Installation failed for pip packages: {failed}")


def setup_environment():
    if git_exe_path is not None:
        git.Git().update_environment(GIT_PYTHON_GIT_EXECUTABLE=git_exe_path)


setup_environment()


try:
    if sys.argv[1] == "--clone":
        repo_path = None
        if len(sys.argv) > 4:
            repo_path = sys.argv[4]

        gitclone(sys.argv[2], sys.argv[3], repo_path=repo_path)
    elif sys.argv[1] == "--check":
        gitcheck(sys.argv[2], False)
    elif sys.argv[1] == "--fetch":
        gitcheck(sys.argv[2], True)
    elif sys.argv[1] == "--pull":
        gitpull(sys.argv[2])
    elif sys.argv[1] == "--apply-snapshot":
        options = set()
        for x in sys.argv:
            if x in ['--pip-non-url', '--pip-local-url', '--pip-non-local-url']:
                options.add(x)

        pips = apply_snapshot(sys.argv[2])

        if pips and len(options) > 0:
            restore_pip_snapshot(pips, options)
    sys.exit(0)
except Exception as e:
    print(e)
    sys.exit(-1)





import concurrent.futures

from comfyui_manager.glob import manager_core as core


def check_state_of_git_node_pack(
    node_packs, do_fetch=False, do_update_check=True, do_update=False
):
    if do_fetch:
        print("Start fetching...", end="")
    elif do_update:
        print("Start updating...", end="")
    elif do_update_check:
        print("Start update check...", end="")

    def process_custom_node(item):
        core.check_state_of_git_node_pack_single(
            item, do_fetch, do_update_check, do_update
        )

    with concurrent.futures.ThreadPoolExecutor(4) as executor:
        for k, v in node_packs.items():
            if v.get("active_version") in ["unknown", "nightly"]:
                executor.submit(process_custom_node, v)

    if do_fetch:
        print("\x1b[2K\rFetching done.")
    elif do_update:
        update_exists = any(
            item.get("updatable", False) for item in node_packs.values()
        )
        if update_exists:
            print("\x1b[2K\rUpdate done.")
        else:
            print("\x1b[2K\rAll extensions are already up-to-date.")
    elif do_update_check:
        print("\x1b[2K\rUpdate check done.")


def nickname_filter(json_obj):
    preemptions_map = {}

    for k, x in json_obj.items():
        if "preemptions" in x[1]:
            for y in x[1]["preemptions"]:
                preemptions_map[y] = k
        elif k.endswith("/ComfyUI"):
            for y in x[0]:
                preemptions_map[y] = k

    updates = {}
    for k, x in json_obj.items():
        removes = set()
        for y in x[0]:
            k2 = preemptions_map.get(y)
            if k2 is not None and k != k2:
                removes.add(y)

        if len(removes) > 0:
            updates[k] = [y for y in x[0] if y not in removes]

    for k, v in updates.items():
        json_obj[k][0] = v

    return json_obj



# generated by datamodel-codegen:
#   filename:  openapi.yaml
#   timestamp: 2025-07-31T04:52:26+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field, RootModel


class OperationType(str, Enum):
    install = "install"
    uninstall = "uninstall"
    update = "update"
    update_comfyui = "update-comfyui"
    fix = "fix"
    disable = "disable"
    enable = "enable"
    install_model = "install-model"


class OperationResult(str, Enum):
    success = "success"
    failed = "failed"
    skipped = "skipped"
    error = "error"
    skip = "skip"


class TaskExecutionStatus(BaseModel):
    status_str: OperationResult
    completed: bool = Field(..., description="Whether the task completed")
    messages: List[str] = Field(..., description="Additional status messages")


class ManagerMessageName(str, Enum):
    cm_task_completed = "cm-task-completed"
    cm_task_started = "cm-task-started"
    cm_queue_status = "cm-queue-status"


class ManagerPackInfo(BaseModel):
    id: str = Field(
        ...,
        description="Either github-author/github-repo or name of pack from the registry",
    )
    version: str = Field(..., description="Semantic version or Git commit hash")
    ui_id: Optional[str] = Field(None, description="Task ID - generated internally")


class ManagerPackInstalled(BaseModel):
    ver: str = Field(
        ...,
        description="The version of the pack that is installed (Git commit hash or semantic version)",
    )
    cnr_id: Optional[str] = Field(
        None, description="The name of the pack if installed from the registry"
    )
    aux_id: Optional[str] = Field(
        None,
        description="The name of the pack if installed from github (author/repo-name format)",
    )
    enabled: bool = Field(..., description="Whether the pack is enabled")


class SelectedVersion(str, Enum):
    latest = "latest"
    nightly = "nightly"


class ManagerChannel(str, Enum):
    default = "default"
    recent = "recent"
    legacy = "legacy"
    forked = "forked"
    dev = "dev"
    tutorial = "tutorial"


class ManagerDatabaseSource(str, Enum):
    remote = "remote"
    local = "local"
    cache = "cache"


class ManagerPackState(str, Enum):
    installed = "installed"
    disabled = "disabled"
    not_installed = "not_installed"
    import_failed = "import_failed"
    needs_update = "needs_update"


class ManagerPackInstallType(str, Enum):
    git_clone = "git-clone"
    copy = "copy"
    cnr = "cnr"


class SecurityLevel(str, Enum):
    strong = "strong"
    normal = "normal"
    normal_ = "normal-"
    weak = "weak"


class RiskLevel(str, Enum):
    block = "block"
    high_ = "high+"
    high = "high"
    middle_ = "middle+"
    middle = "middle"


class UpdateState(Enum):
    false = "false"
    true = "true"


class ManagerPack(ManagerPackInfo):
    author: Optional[str] = Field(
        None, description="Pack author name or 'Unclaimed' if added via GitHub crawl"
    )
    files: Optional[List[str]] = Field(
        None,
        description="Repository URLs for installation (typically contains one GitHub URL)",
    )
    reference: Optional[str] = Field(
        None, description="The type of installation reference"
    )
    title: Optional[str] = Field(None, description="The display name of the pack")
    cnr_latest: Optional[SelectedVersion] = None
    repository: Optional[str] = Field(None, description="GitHub repository URL")
    state: Optional[ManagerPackState] = None
    update_state: Optional[UpdateState] = Field(
        None, alias="update-state", description="Update availability status"
    )
    stars: Optional[int] = Field(None, description="GitHub stars count")
    last_update: Optional[datetime] = Field(None, description="Last update timestamp")
    health: Optional[str] = Field(None, description="Health status of the pack")
    description: Optional[str] = Field(None, description="Pack description")
    trust: Optional[bool] = Field(None, description="Whether the pack is trusted")
    install_type: Optional[ManagerPackInstallType] = None


class InstallPackParams(ManagerPackInfo):
    selected_version: Union[str, SelectedVersion] = Field(
        ..., description="Semantic version, Git commit hash, latest, or nightly"
    )
    repository: Optional[str] = Field(
        None,
        description="GitHub repository URL (required if selected_version is nightly)",
    )
    pip: Optional[List[str]] = Field(None, description="PyPi dependency names")
    mode: ManagerDatabaseSource
    channel: ManagerChannel
    skip_post_install: Optional[bool] = Field(
        None, description="Whether to skip post-installation steps"
    )


class UpdateAllPacksParams(BaseModel):
    mode: Optional[ManagerDatabaseSource] = None
    ui_id: Optional[str] = Field(None, description="Task ID - generated internally")


class UpdatePackParams(BaseModel):
    node_name: str = Field(..., description="Name of the node package to update")
    node_ver: Optional[str] = Field(
        None, description="Current version of the node package"
    )


class UpdateComfyUIParams(BaseModel):
    is_stable: Optional[bool] = Field(
        True,
        description="Whether to update to stable version (true) or nightly (false)",
    )
    target_version: Optional[str] = Field(
        None,
        description="Specific version to switch to (for version switching operations)",
    )


class FixPackParams(BaseModel):
    node_name: str = Field(..., description="Name of the node package to fix")
    node_ver: str = Field(..., description="Version of the node package")


class UninstallPackParams(BaseModel):
    node_name: str = Field(..., description="Name of the node package to uninstall")
    is_unknown: Optional[bool] = Field(
        False, description="Whether this is an unknown/unregistered package"
    )


class DisablePackParams(BaseModel):
    node_name: str = Field(..., description="Name of the node package to disable")
    is_unknown: Optional[bool] = Field(
        False, description="Whether this is an unknown/unregistered package"
    )


class EnablePackParams(BaseModel):
    cnr_id: str = Field(
        ..., description="ComfyUI Node Registry ID of the package to enable"
    )


class UpdateAllQueryParams(BaseModel):
    client_id: str = Field(
        ..., description="Client identifier that initiated the request"
    )
    ui_id: str = Field(..., description="Base UI identifier for task tracking")
    mode: Optional[ManagerDatabaseSource] = None


class UpdateComfyUIQueryParams(BaseModel):
    client_id: str = Field(
        ..., description="Client identifier that initiated the request"
    )
    ui_id: str = Field(..., description="UI identifier for task tracking")
    stable: Optional[bool] = Field(
        True,
        description="Whether to update to stable version (true) or nightly (false)",
    )


class ComfyUISwitchVersionQueryParams(BaseModel):
    ver: str = Field(..., description="Version to switch to")
    client_id: str = Field(
        ..., description="Client identifier that initiated the request"
    )
    ui_id: str = Field(..., description="UI identifier for task tracking")


class QueueStatus(BaseModel):
    total_count: int = Field(
        ..., description="Total number of tasks (pending + running)"
    )
    done_count: int = Field(..., description="Number of completed tasks")
    in_progress_count: int = Field(..., description="Number of tasks currently running")
    pending_count: Optional[int] = Field(
        None, description="Number of tasks waiting to be executed"
    )
    is_processing: bool = Field(..., description="Whether the task worker is active")
    client_id: Optional[str] = Field(
        None, description="Client ID (when filtered by client)"
    )


class ManagerMappings1(BaseModel):
    title_aux: Optional[str] = Field(None, description="The display name of the pack")


class ManagerMappings(
    RootModel[Optional[Dict[str, List[Union[List[str], ManagerMappings1]]]]]
):
    root: Optional[Dict[str, List[Union[List[str], ManagerMappings1]]]] = Field(
        None, description="Tuple of [node_names, metadata]"
    )


class ModelMetadata(BaseModel):
    name: str = Field(..., description="Name of the model")
    type: str = Field(..., description="Type of model")
    base: Optional[str] = Field(None, description="Base model type")
    save_path: Optional[str] = Field(None, description="Path for saving the model")
    url: str = Field(..., description="Download URL")
    filename: str = Field(..., description="Target filename")
    ui_id: Optional[str] = Field(None, description="ID for UI reference")


class InstallType(str, Enum):
    git = "git"
    copy = "copy"
    pip = "pip"


class NodePackageMetadata(BaseModel):
    title: Optional[str] = Field(None, description="Display name of the node package")
    name: Optional[str] = Field(None, description="Repository/package name")
    files: Optional[List[str]] = Field(None, description="Source URLs for the package")
    description: Optional[str] = Field(
        None, description="Description of the node package functionality"
    )
    install_type: Optional[InstallType] = Field(None, description="Installation method")
    version: Optional[str] = Field(None, description="Version identifier")
    id: Optional[str] = Field(
        None, description="Unique identifier for the node package"
    )
    ui_id: Optional[str] = Field(None, description="ID for UI reference")
    channel: Optional[str] = Field(None, description="Source channel")
    mode: Optional[str] = Field(None, description="Source mode")


class SnapshotItem(RootModel[str]):
    root: str = Field(..., description="Name of the snapshot")


class Error(BaseModel):
    error: str = Field(..., description="Error message")


class InstalledPacksResponse(RootModel[Optional[Dict[str, ManagerPackInstalled]]]):
    root: Optional[Dict[str, ManagerPackInstalled]] = None


class HistoryListResponse(BaseModel):
    ids: Optional[List[str]] = Field(
        None, description="List of available batch history IDs"
    )


class InstalledNodeInfo(BaseModel):
    name: str = Field(..., description="Node package name")
    version: str = Field(..., description="Installed version")
    repository_url: Optional[str] = Field(None, description="Git repository URL")
    install_method: str = Field(
        ..., description="Installation method (cnr, git, pip, etc.)"
    )
    enabled: Optional[bool] = Field(
        True, description="Whether the node is currently enabled"
    )
    install_date: Optional[datetime] = Field(
        None, description="ISO timestamp of installation"
    )


class InstalledModelInfo(BaseModel):
    name: str = Field(..., description="Model filename")
    path: str = Field(..., description="Full path to model file")
    type: str = Field(..., description="Model type (checkpoint, lora, vae, etc.)")
    size_bytes: Optional[int] = Field(None, description="File size in bytes", ge=0)
    hash: Optional[str] = Field(None, description="Model file hash for verification")
    install_date: Optional[datetime] = Field(
        None, description="ISO timestamp when added"
    )


class ComfyUIVersionInfo(BaseModel):
    version: str = Field(..., description="ComfyUI version string")
    commit_hash: Optional[str] = Field(None, description="Git commit hash")
    branch: Optional[str] = Field(None, description="Git branch name")
    is_stable: Optional[bool] = Field(
        False, description="Whether this is a stable release"
    )
    last_updated: Optional[datetime] = Field(
        None, description="ISO timestamp of last update"
    )


class BatchOperation(BaseModel):
    operation_id: str = Field(..., description="Unique operation identifier")
    operation_type: OperationType
    target: str = Field(
        ..., description="Target of the operation (node name, model name, etc.)"
    )
    target_version: Optional[str] = Field(
        None, description="Target version for the operation"
    )
    result: OperationResult
    error_message: Optional[str] = Field(
        None, description="Error message if operation failed"
    )
    start_time: datetime = Field(
        ..., description="ISO timestamp when operation started"
    )
    end_time: Optional[datetime] = Field(
        None, description="ISO timestamp when operation completed"
    )
    client_id: Optional[str] = Field(
        None, description="Client that initiated the operation"
    )


class ComfyUISystemState(BaseModel):
    snapshot_time: datetime = Field(
        ..., description="ISO timestamp when snapshot was taken"
    )
    comfyui_version: ComfyUIVersionInfo
    frontend_version: Optional[str] = Field(
        None, description="ComfyUI frontend version if available"
    )
    python_version: str = Field(..., description="Python interpreter version")
    platform_info: str = Field(
        ..., description="Operating system and platform information"
    )
    installed_nodes: Optional[Dict[str, InstalledNodeInfo]] = Field(
        None, description="Map of installed node packages by name"
    )
    installed_models: Optional[Dict[str, InstalledModelInfo]] = Field(
        None, description="Map of installed models by name"
    )
    manager_config: Optional[Dict[str, Any]] = Field(
        None, description="ComfyUI Manager configuration settings"
    )
    comfyui_root_path: Optional[str] = Field(
        None, description="ComfyUI root installation directory"
    )
    model_paths: Optional[Dict[str, List[str]]] = Field(
        None, description="Map of model types to their configured paths"
    )
    manager_version: Optional[str] = Field(None, description="ComfyUI Manager version")
    security_level: Optional[SecurityLevel] = None
    network_mode: Optional[str] = Field(
        None, description="Network mode (online, offline, private)"
    )
    cli_args: Optional[Dict[str, Any]] = Field(
        None, description="Selected ComfyUI CLI arguments"
    )
    custom_nodes_count: Optional[int] = Field(
        None, description="Total number of custom node packages", ge=0
    )
    failed_imports: Optional[List[str]] = Field(
        None, description="List of custom nodes that failed to import"
    )
    pip_packages: Optional[Dict[str, str]] = Field(
        None, description="Map of installed pip packages to their versions"
    )
    embedded_python: Optional[bool] = Field(
        None,
        description="Whether ComfyUI is running from an embedded Python distribution",
    )


class BatchExecutionRecord(BaseModel):
    batch_id: str = Field(..., description="Unique batch identifier")
    start_time: datetime = Field(..., description="ISO timestamp when batch started")
    end_time: Optional[datetime] = Field(
        None, description="ISO timestamp when batch completed"
    )
    state_before: ComfyUISystemState
    state_after: Optional[ComfyUISystemState] = Field(
        None, description="System state after batch execution"
    )
    operations: Optional[List[BatchOperation]] = Field(
        None, description="List of operations performed in this batch"
    )
    total_operations: Optional[int] = Field(
        0, description="Total number of operations in batch", ge=0
    )
    successful_operations: Optional[int] = Field(
        0, description="Number of successful operations", ge=0
    )
    failed_operations: Optional[int] = Field(
        0, description="Number of failed operations", ge=0
    )
    skipped_operations: Optional[int] = Field(
        0, description="Number of skipped operations", ge=0
    )


class ImportFailInfoBulkRequest(BaseModel):
    cnr_ids: Optional[List[str]] = Field(
        None, description="A list of CNR IDs to check."
    )
    urls: Optional[List[str]] = Field(
        None, description="A list of repository URLs to check."
    )


class ImportFailInfoItem1(BaseModel):
    error: Optional[str] = None
    traceback: Optional[str] = None


class ImportFailInfoItem(RootModel[Optional[ImportFailInfoItem1]]):
    root: Optional[ImportFailInfoItem1]


class QueueTaskItem(BaseModel):
    ui_id: str = Field(..., description="Unique identifier for the task")
    client_id: str = Field(..., description="Client identifier that initiated the task")
    kind: OperationType
    params: Union[
        InstallPackParams,
        UpdatePackParams,
        UpdateAllPacksParams,
        UpdateComfyUIParams,
        FixPackParams,
        UninstallPackParams,
        DisablePackParams,
        EnablePackParams,
        ModelMetadata,
    ]


class TaskHistoryItem(BaseModel):
    ui_id: str = Field(..., description="Unique identifier for the task")
    client_id: str = Field(..., description="Client identifier that initiated the task")
    kind: str = Field(..., description="Type of task that was performed")
    timestamp: datetime = Field(..., description="ISO timestamp when task completed")
    result: str = Field(..., description="Task result message or details")
    status: Optional[TaskExecutionStatus] = None
    batch_id: Optional[str] = Field(
        None, description="ID of the batch this task belongs to"
    )
    end_time: Optional[datetime] = Field(
        None, description="ISO timestamp when task execution ended"
    )


class TaskStateMessage(BaseModel):
    history: Dict[str, TaskHistoryItem] = Field(
        ..., description="Map of task IDs to their history items"
    )
    running_queue: List[QueueTaskItem] = Field(
        ..., description="Currently executing tasks"
    )
    pending_queue: List[QueueTaskItem] = Field(
        ..., description="Tasks waiting to be executed"
    )
    installed_packs: Dict[str, ManagerPackInstalled] = Field(
        ..., description="Map of currently installed node packages by name"
    )


class MessageTaskDone(BaseModel):
    ui_id: str = Field(..., description="Task identifier")
    result: str = Field(..., description="Task result message")
    kind: str = Field(..., description="Type of task")
    status: Optional[TaskExecutionStatus] = None
    timestamp: datetime = Field(..., description="ISO timestamp when task completed")
    state: TaskStateMessage


class MessageTaskStarted(BaseModel):
    ui_id: str = Field(..., description="Task identifier")
    kind: str = Field(..., description="Type of task")
    timestamp: datetime = Field(..., description="ISO timestamp when task started")
    state: TaskStateMessage


class MessageTaskFailed(BaseModel):
    ui_id: str = Field(..., description="Task identifier")
    error: str = Field(..., description="Error message")
    kind: str = Field(..., description="Type of task")
    timestamp: datetime = Field(..., description="ISO timestamp when task failed")
    state: TaskStateMessage


class MessageUpdate(
    RootModel[Union[MessageTaskDone, MessageTaskStarted, MessageTaskFailed]]
):
    root: Union[MessageTaskDone, MessageTaskStarted, MessageTaskFailed] = Field(
        ..., description="Union type for all possible WebSocket message updates"
    )


class HistoryResponse(BaseModel):
    history: Optional[Dict[str, TaskHistoryItem]] = Field(
        None, description="Map of task IDs to their history items"
    )


class ImportFailInfoBulkResponse(RootModel[Optional[Dict[str, ImportFailInfoItem]]]):
    root: Optional[Dict[str, ImportFailInfoItem]] = None



import asyncio
import json
import os
import platform
import time
from dataclasses import dataclass
from typing import List

from . import context
from . import manager_util

import requests
import toml

base_url = "https://api.comfy.org"


lock = asyncio.Lock()

is_cache_loading = False

async def get_cnr_data(cache_mode=True, dont_wait=True):
    try:
        return await _get_cnr_data(cache_mode, dont_wait)
    except asyncio.TimeoutError:
        print("A timeout occurred during the fetch process from ComfyRegistry.")
        return await _get_cnr_data(cache_mode=True, dont_wait=True)  # timeout fallback

async def _get_cnr_data(cache_mode=True, dont_wait=True):
    global is_cache_loading

    uri = f'{base_url}/nodes'

    async def fetch_all():
        remained = True
        page = 1

        full_nodes = {}

        
        # Determine form factor based on environment and platform
        is_desktop = bool(os.environ.get('__COMFYUI_DESKTOP_VERSION__'))
        system = platform.system().lower()
        is_windows = system == 'windows'
        is_mac = system == 'darwin'
        is_linux = system == 'linux'

        # Get ComfyUI version tag
        if is_desktop:
            # extract version from pyproject.toml instead of git tag
            comfyui_ver = context.get_current_comfyui_ver() or 'unknown'
        else:
            comfyui_ver = context.get_comfyui_tag() or 'unknown'

        if is_desktop:
            if is_windows:
                form_factor = 'desktop-win'
            elif is_mac:
                form_factor = 'desktop-mac'
            else:
                form_factor = 'other'
        else:
            if is_windows:
                form_factor = 'git-windows'
            elif is_mac:
                form_factor = 'git-mac'
            elif is_linux:
                form_factor = 'git-linux'
            else:
                form_factor = 'other'
        
        while remained:
            # Add comfyui_version and form_factor to the API request
            sub_uri = f'{base_url}/nodes?page={page}&limit=30&comfyui_version={comfyui_ver}&form_factor={form_factor}'
            sub_json_obj = await asyncio.wait_for(manager_util.get_data_with_cache(sub_uri, cache_mode=False, silent=True, dont_cache=True), timeout=30)
            remained = page < sub_json_obj['totalPages']

            for x in sub_json_obj['nodes']:
                full_nodes[x['id']] = x

            if page % 5 == 0:
                print(f"FETCH ComfyRegistry Data: {page}/{sub_json_obj['totalPages']}")

            page += 1
            time.sleep(0.5)

        print("FETCH ComfyRegistry Data [DONE]")

        for v in full_nodes.values():
            if 'latest_version' not in v:
                v['latest_version'] = dict(version='nightly')

        return {'nodes': list(full_nodes.values())}

    if cache_mode:
        is_cache_loading = True
        cache_state = manager_util.get_cache_state(uri)

        if dont_wait:
            if cache_state == 'not-cached':
                return {}
            else:
                print("[ComfyUI-Manager] The ComfyRegistry cache update is still in progress, so an outdated cache is being used.")
                with open(manager_util.get_cache_path(uri), 'r', encoding="UTF-8", errors="ignore") as json_file:
                    return json.load(json_file)['nodes']

        if cache_state == 'cached':
            with open(manager_util.get_cache_path(uri), 'r', encoding="UTF-8", errors="ignore") as json_file:
                return json.load(json_file)['nodes']

    try:
        json_obj = await fetch_all()
        manager_util.save_to_cache(uri, json_obj)
        return json_obj['nodes']
    except Exception:
        res = {}
        print("Cannot connect to comfyregistry.")
    finally:
        if cache_mode:
            is_cache_loading = False

    return res


@dataclass
class NodeVersion:
    changelog: str
    dependencies: List[str]
    deprecated: bool
    id: str
    version: str
    download_url: str


def map_node_version(api_node_version):
    """
    Maps node version data from API response to NodeVersion dataclass.

    Args:
        api_data (dict): The 'node_version' part of the API response.

    Returns:
        NodeVersion: An instance of NodeVersion dataclass populated with data from the API.
    """
    return NodeVersion(
        changelog=api_node_version.get(
            "changelog", ""
        ),  # Provide a default value if 'changelog' is missing
        dependencies=api_node_version.get(
            "dependencies", []
        ),  # Provide a default empty list if 'dependencies' is missing
        deprecated=api_node_version.get(
            "deprecated", False
        ),  # Assume False if 'deprecated' is not specified
        id=api_node_version[
            "id"
        ],  # 'id' should be mandatory; raise KeyError if missing
        version=api_node_version[
            "version"
        ],  # 'version' should be mandatory; raise KeyError if missing
        download_url=api_node_version.get(
            "downloadUrl", ""
        ),  # Provide a default value if 'downloadUrl' is missing
    )


def install_node(node_id, version=None):
    """
    Retrieves the node version for installation.

    Args:
      node_id (str): The unique identifier of the node.
      version (str, optional): Specific version of the node to retrieve. If omitted, the latest version is returned.

    Returns:
      NodeVersion: Node version data or error message.
    """
    if version is None:
        url = f"{base_url}/nodes/{node_id}/install"
    else:
        url = f"{base_url}/nodes/{node_id}/install?version={version}"

    response = requests.get(url, verify=not manager_util.bypass_ssl)
    if response.status_code == 200:
        # Convert the API response to a NodeVersion object
        return map_node_version(response.json())
    else:
        return None


def all_versions_of_node(node_id):
    url = f"{base_url}/nodes/{node_id}/versions?statuses=NodeVersionStatusActive&statuses=NodeVersionStatusPending"

    response = requests.get(url, verify=not manager_util.bypass_ssl)
    if response.status_code == 200:
        return response.json()
    else:
        return None


def read_cnr_info(fullpath):
    try:
        toml_path = os.path.join(fullpath, 'pyproject.toml')
        tracking_path = os.path.join(fullpath, '.tracking')

        if not os.path.exists(toml_path) or not os.path.exists(tracking_path):
            return None  # not valid CNR node pack

        with open(toml_path, "r", encoding="utf-8") as f:
            data = toml.load(f)

            project = data.get('project', {})
            name = project.get('name').strip().lower()

            # normalize version
            # for example: 2.5 -> 2.5.0
            version = str(manager_util.StrictVersion(project.get('version')))

            urls = project.get('urls', {})
            repository = urls.get('Repository')

            if name and version:  # repository is optional
                return {
                    "id": name,
                    "version": version,
                    "url": repository
                }

        return None
    except Exception:
        return None  # not valid CNR node pack


def generate_cnr_id(fullpath, cnr_id):
    cnr_id_path = os.path.join(fullpath, '.git', '.cnr-id')
    try:
        if not os.path.exists(cnr_id_path):
            with open(cnr_id_path, "w") as f:
                return f.write(cnr_id)
    except Exception:
        print(f"[ComfyUI Manager] unable to create file: {cnr_id_path}")


def read_cnr_id(fullpath):
    cnr_id_path = os.path.join(fullpath, '.git', '.cnr-id')
    try:
        if os.path.exists(cnr_id_path):
            with open(cnr_id_path) as f:
                return f.read().strip()
    except Exception:
        pass

    return None




import enum

class NetworkMode(enum.Enum):
    PUBLIC = "public"
    PRIVATE = "private"
    OFFLINE = "offline"
    PERSONAL_CLOUD = "personal_cloud"

class SecurityLevel(enum.Enum):
    STRONG = "strong"
    NORMAL = "normal"
    NORMAL_MINUS = "normal-minus"
    WEAK = "weak"

class DBMode(enum.Enum):
    LOCAL = "local"
    CACHE = "cache"
    REMOTE = "remote"